Type: Package
Package: metatools
Title: Enable the Use of 'metacore' to Help Create and Check Dataset
Version: 0.3.0
Authors@R: c(
    person("Liam", "Hobby", , "liam.f.hobby@gsk.com", role = c("aut", "cre")),
    person("Christina", "Fillmore", , "christina.e.fillmore@gsk.com", role = "aut",
           comment = c(ORCID = "0000-0003-0595-2302")),
    person("Bill", "Denney", role = "aut"),
    person("Mike", "Stackhouse", , "mike.stackhouse@atorusresearch.com", role = "aut",
           comment = c(ORCID = "0000-0001-6030-723X")),
    person("Jana", "Stoilova", , "jana.stoilova@roche.com", role = "aut"),
    person("Tamara", "Senior", , "tamara.senior@roche.com", role = "aut"),
    person("GlaxoSmithKline LLC", role = c("cph", "fnd")),
    person("F. Hoffmann-La Roche AG", role = c("cph", "fnd")),
    person("Atorus Research LLC", role = c("cph", "fnd"))
  )
Description: Uses the metadata information stored in 'metacore' objects to
    check and build metadata associated columns.
License: MIT + file LICENSE
URL: https://github.com/pharmaverse/metatools,
        https://pharmaverse.github.io/metatools/
BugReports: https://github.com/pharmaverse/metatools/issues
Depends: R (>= 4.1.0)
Imports: cli, dplyr, lifecycle, magrittr, metacore (>= 0.2.0), purrr,
        rlang, stringr, tibble, tidyr
Suggests: covr, haven, pharmaversesdtm, safetyData, spelling, testthat
        (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-28 11:12:23 UTC; lfh66519
Author: Liam Hobby [aut, cre],
  Christina Fillmore [aut] (ORCID:
    <https://orcid.org/0000-0003-0595-2302>),
  Bill Denney [aut],
  Mike Stackhouse [aut] (ORCID: <https://orcid.org/0000-0001-6030-723X>),
  Jana Stoilova [aut],
  Tamara Senior [aut],
  GlaxoSmithKline LLC [cph, fnd],
  F. Hoffmann-La Roche AG [cph, fnd],
  Atorus Research LLC [cph, fnd]
Maintainer: Liam Hobby <liam.f.hobby@gsk.com>
Repository: CRAN
Date/Publication: 2026-04-28 11:30:02 UTC
Built: R 4.5.2; ; 2026-04-28 15:34:32 UTC; unix

Type: Package
Package: metatools
Title: Enable the Use of 'metacore' to Help Create and Check Dataset
Version: 0.3.0
Authors@R: c(
    person("Liam", "Hobby", , "liam.f.hobby@gsk.com", role = c("aut", "cre")),
    person("Christina", "Fillmore", , "christina.e.fillmore@gsk.com", role = "aut",
           comment = c(ORCID = "0000-0003-0595-2302")),
    person("Bill", "Denney", role = "aut"),
    person("Mike", "Stackhouse", , "mike.stackhouse@atorusresearch.com", role = "aut",
           comment = c(ORCID = "0000-0001-6030-723X")),
    person("Jana", "Stoilova", , "jana.stoilova@roche.com", role = "aut"),
    person("Tamara", "Senior", , "tamara.senior@roche.com", role = "aut"),
    person("GlaxoSmithKline LLC", role = c("cph", "fnd")),
    person("F. Hoffmann-La Roche AG", role = c("cph", "fnd")),
    person("Atorus Research LLC", role = c("cph", "fnd"))
  )
Description: Uses the metadata information stored in 'metacore' objects to
    check and build metadata associated columns.
License: MIT + file LICENSE
URL: https://github.com/pharmaverse/metatools,
        https://pharmaverse.github.io/metatools/
BugReports: https://github.com/pharmaverse/metatools/issues
Depends: R (>= 4.1.0)
Imports: cli, dplyr, lifecycle, magrittr, metacore (>= 0.2.0), purrr,
        rlang, stringr, tibble, tidyr
Suggests: covr, haven, pharmaversesdtm, safetyData, spelling, testthat
        (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-28 11:12:23 UTC; lfh66519
Author: Liam Hobby [aut, cre],
  Christina Fillmore [aut] (ORCID:
    <https://orcid.org/0000-0003-0595-2302>),
  Bill Denney [aut],
  Mike Stackhouse [aut] (ORCID: <https://orcid.org/0000-0001-6030-723X>),
  Jana Stoilova [aut],
  Tamara Senior [aut],
  GlaxoSmithKline LLC [cph, fnd],
  F. Hoffmann-La Roche AG [cph, fnd],
  Atorus Research LLC [cph, fnd]
Maintainer: Liam Hobby <liam.f.hobby@gsk.com>
Repository: CRAN
Date/Publication: 2026-04-28 11:30:02 UTC
Built: R 4.6.0; ; 2026-04-28 15:48:27 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(add_labels)
export(add_variables)
export(build_from_derived)
export(build_qnam)
export(check_ct_col)
export(check_ct_data)
export(check_unique_keys)
export(check_variables)
export(combine_supp)
export(convert_var_to_fct)
export(create_cat_var)
export(create_subgrps)
export(create_var_from_codelist)
export(drop_unspec_vars)
export(get_bad_ct)
export(make_supp_qual)
export(metatools_example)
export(order_cols)
export(remove_labels)
export(set_variable_labels)
export(sort_by_key)
importFrom(cli,ansi_collapse)
importFrom(cli,cli_abort)
importFrom(cli,cli_inform)
importFrom(cli,cli_text)
importFrom(cli,cli_warn)
importFrom(cli,col_red)
importFrom(cli,qty)
importFrom(dplyr,across)
importFrom(dplyr,add_count)
importFrom(dplyr,all_of)
importFrom(dplyr,anti_join)
importFrom(dplyr,any_of)
importFrom(dplyr,arrange)
importFrom(dplyr,as_tibble)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,case_when)
importFrom(dplyr,distinct)
importFrom(dplyr,enquo)
importFrom(dplyr,everything)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,group_split)
importFrom(dplyr,if_else)
importFrom(dplyr,inner_join)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,pick)
importFrom(dplyr,pull)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(lifecycle,deprecate_warn)
importFrom(lifecycle,deprecated)
importFrom(lifecycle,is_present)
importFrom(magrittr,"%>%")
importFrom(metacore,get_control_term)
importFrom(metacore,get_keys)
importFrom(metacore,select_dataset)
importFrom(metacore,verify_DatasetMeta)
importFrom(purrr,discard)
importFrom(purrr,keep)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_chr)
importFrom(purrr,map_dfr)
importFrom(purrr,map_lgl)
importFrom(purrr,pmap_dfr)
importFrom(purrr,reduce)
importFrom(purrr,safely)
importFrom(purrr,walk2)
importFrom(rlang,"!!")
importFrom(rlang,":=")
importFrom(rlang,as_label)
importFrom(rlang,as_name)
importFrom(rlang,enexpr)
importFrom(rlang,expr)
importFrom(rlang,is_logical)
importFrom(rlang,is_named)
importFrom(rlang,list2)
importFrom(rlang,set_names)
importFrom(rlang,sym)
importFrom(stringr,str_c)
importFrom(stringr,str_detect)
importFrom(stringr,str_extract)
importFrom(stringr,str_match_all)
importFrom(stringr,str_remove)
importFrom(stringr,str_remove_all)
importFrom(stringr,str_split)
importFrom(stringr,str_to_lower)
importFrom(stringr,str_to_upper)
importFrom(stringr,str_trim)
importFrom(tibble,tibble)
importFrom(tidyr,pivot_wider)
importFrom(tidyr,unnest)
importFrom(utils,capture.output)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# metatools <a href='https://github.com/pharmaverse/metatools'><img src="man/figures/metatools.png" align="right" style="height:139px;"/></a>

<!-- badges: start -->

[<img src="http://pharmaverse.org/shields/metatools.svg"/>](https://pharmaverse.org)
[![R-CMD-check](https://github.com/pharmaverse/metatools/workflows/R-CMD-check/badge.svg)](https://github.com/pharmaverse/metatools/actions)
[![codecov](https://codecov.io/gh/pharmaverse/metatools/branch/main/graph/badge.svg?token=55N5APFLPA)](https://app.codecov.io/gh/pharmaverse/metatools)
[<img src="https://img.shields.io/badge/License-MIT-blue.svg"/>](https://github.com/pharmaverse/metatools/blob/main/LICENSE)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/pharmaverse/metatools/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/pharmaverse/metatools/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

The goal of metatools is to enable the use of metacore objects.
Metatools can be used to build datasets or enhance columns in existing
datasets as well as checking datasets against the metadata in metacore.

## Installation

You can install the latest released version of metatools from CRAN with:

``` r
install.packages("metatools")
```

And the current development version from
[GitHub](https://github.com/pharmaverse/metatools) with:

``` r
devtools::install_github("pharmaverse/metatools")
```

## Example

Here is a basic example of some functions to build datasets and create
new columns.

``` r
library(metatools)
library(metacore)
library(haven)
library(dplyr)
#> 
#> Attaching package: 'dplyr'
#> The following objects are masked from 'package:stats':
#> 
#>     filter, lag
#> The following objects are masked from 'package:base':
#> 
#>     intersect, setdiff, setequal, union
load(metacore_example("pilot_ADaM.rda"))
metacore <- metacore %>%
  select_dataset("ADSL")
ds_list <- list(DM = read_xpt(metatools_example("dm.xpt")))
# Pull in columns from DM to be in ADSL
adsl <- build_from_derived(metacore, ds_list, predecessor_only = FALSE) 
adsl 
#> # A tibble: 306 × 13
#>    STUDYID     USUBJID SUBJID SITEID TRT01P   AGE AGEU  RACE  SEX   ETHNIC DTHFL
#>    <chr>       <chr>   <chr>  <chr>  <chr>  <dbl> <chr> <chr> <chr> <chr>  <chr>
#>  1 CDISCPILOT… 01-701… 1015   701    Place…    63 YEARS WHITE F     HISPA… ""   
#>  2 CDISCPILOT… 01-701… 1023   701    Place…    64 YEARS WHITE M     HISPA… ""   
#>  3 CDISCPILOT… 01-701… 1028   701    Xanom…    71 YEARS WHITE M     NOT H… ""   
#>  4 CDISCPILOT… 01-701… 1033   701    Xanom…    74 YEARS WHITE M     NOT H… ""   
#>  5 CDISCPILOT… 01-701… 1034   701    Xanom…    77 YEARS WHITE F     NOT H… ""   
#>  6 CDISCPILOT… 01-701… 1047   701    Place…    85 YEARS WHITE F     NOT H… ""   
#>  7 CDISCPILOT… 01-701… 1057   701    Scree…    59 YEARS WHITE F     HISPA… ""   
#>  8 CDISCPILOT… 01-701… 1097   701    Xanom…    68 YEARS WHITE M     NOT H… ""   
#>  9 CDISCPILOT… 01-701… 1111   701    Xanom…    81 YEARS WHITE F     NOT H… ""   
#> 10 CDISCPILOT… 01-701… 1115   701    Xanom…    84 YEARS WHITE M     NOT H… ""   
#> # … with 296 more rows, and 2 more variables: RFSTDTC <chr>, RFENDTC <chr>

# Now we can make some new columns and converting columns into factors
adsl %>%
   select(USUBJID, AGE, ETHNIC) %>% 
   create_cat_var(metacore, AGE, AGEGR1, AGEGR1N) %>% #Add an AGEGR1 and AGEGR1N column
   convert_var_to_fct(metacore, ETHNIC) # Change ETHNIC to as factor 
#> # A tibble: 306 × 5
#>    USUBJID       AGE ETHNIC                 AGEGR1 AGEGR1N
#>    <chr>       <dbl> <fct>                  <chr>    <dbl>
#>  1 01-701-1015    63 HISPANIC OR LATINO     <65          1
#>  2 01-701-1023    64 HISPANIC OR LATINO     <65          1
#>  3 01-701-1028    71 NOT HISPANIC OR LATINO 65-80        2
#>  4 01-701-1033    74 NOT HISPANIC OR LATINO 65-80        2
#>  5 01-701-1034    77 NOT HISPANIC OR LATINO 65-80        2
#>  6 01-701-1047    85 NOT HISPANIC OR LATINO >80          3
#>  7 01-701-1057    59 HISPANIC OR LATINO     <65          1
#>  8 01-701-1097    68 NOT HISPANIC OR LATINO 65-80        2
#>  9 01-701-1111    81 NOT HISPANIC OR LATINO >80          3
#> 10 01-701-1115    84 NOT HISPANIC OR LATINO >80          3
#> # … with 296 more rows
```

Metatools can also be used to run checks

``` r
data <- read_xpt(metatools_example("adsl.xpt"))

# Checks can be run on a single column
check_ct_col(data, metacore, TRT01PN) # Checks column only contains control terminology
#> # A tibble: 254 × 49
#>    STUDYID     USUBJID SUBJID SITEID SITEGR1 ARM   TRT01P TRT01PN TRT01A TRT01AN
#>    <chr>       <chr>   <chr>  <chr>  <chr>   <chr> <chr>    <dbl> <chr>    <dbl>
#>  1 CDISCPILOT… 01-701… 1015   701    701     Plac… Place…       0 Place…       0
#>  2 CDISCPILOT… 01-701… 1023   701    701     Plac… Place…       0 Place…       0
#>  3 CDISCPILOT… 01-701… 1028   701    701     Xano… Xanom…      81 Xanom…      81
#>  4 CDISCPILOT… 01-701… 1033   701    701     Xano… Xanom…      54 Xanom…      54
#>  5 CDISCPILOT… 01-701… 1034   701    701     Xano… Xanom…      81 Xanom…      81
#>  6 CDISCPILOT… 01-701… 1047   701    701     Plac… Place…       0 Place…       0
#>  7 CDISCPILOT… 01-701… 1097   701    701     Xano… Xanom…      54 Xanom…      54
#>  8 CDISCPILOT… 01-701… 1111   701    701     Xano… Xanom…      54 Xanom…      54
#>  9 CDISCPILOT… 01-701… 1115   701    701     Xano… Xanom…      54 Xanom…      54
#> 10 CDISCPILOT… 01-701… 1118   701    701     Plac… Place…       0 Place…       0
#> # … with 244 more rows, and 39 more variables: TRTSDT <date>, TRTEDT <date>,
#> #   TRTDURD <dbl>, AVGDD <dbl>, CUMDOSE <dbl>, AGE <dbl>, AGEGR1 <chr>,
#> #   AGEGR1N <dbl>, AGEU <chr>, RACE <chr>, RACEN <dbl>, SEX <chr>,
#> #   ETHNIC <chr>, SAFFL <chr>, ITTFL <chr>, EFFFL <chr>, COMP8FL <chr>,
#> #   COMP16FL <chr>, COMP24FL <chr>, DISCONFL <chr>, DSRAEFL <chr>, DTHFL <chr>,
#> #   BMIBL <dbl>, BMIBLGR1 <chr>, HEIGHTBL <dbl>, WEIGHTBL <dbl>, EDUCLVL <dbl>,
#> #   DISONSDT <date>, DURDIS <dbl>, DURDSGR1 <chr>, VISIT1DT <date>, …
# Or across all the columns 
check_ct_data(data, metacore) %>%  # Checks control terminology for all columns 
check_variables(metacore) # Check all variables in the metadata are in the dataset and there aren't any extra columns 
#> No missing or extra variables
#> # A tibble: 254 × 49
#>    STUDYID     USUBJID SUBJID SITEID SITEGR1 ARM   TRT01P TRT01PN TRT01A TRT01AN
#>    <chr>       <chr>   <chr>  <chr>  <chr>   <chr> <chr>    <dbl> <chr>    <dbl>
#>  1 CDISCPILOT… 01-701… 1015   701    701     Plac… Place…       0 Place…       0
#>  2 CDISCPILOT… 01-701… 1023   701    701     Plac… Place…       0 Place…       0
#>  3 CDISCPILOT… 01-701… 1028   701    701     Xano… Xanom…      81 Xanom…      81
#>  4 CDISCPILOT… 01-701… 1033   701    701     Xano… Xanom…      54 Xanom…      54
#>  5 CDISCPILOT… 01-701… 1034   701    701     Xano… Xanom…      81 Xanom…      81
#>  6 CDISCPILOT… 01-701… 1047   701    701     Plac… Place…       0 Place…       0
#>  7 CDISCPILOT… 01-701… 1097   701    701     Xano… Xanom…      54 Xanom…      54
#>  8 CDISCPILOT… 01-701… 1111   701    701     Xano… Xanom…      54 Xanom…      54
#>  9 CDISCPILOT… 01-701… 1115   701    701     Xano… Xanom…      54 Xanom…      54
#> 10 CDISCPILOT… 01-701… 1118   701    701     Plac… Place…       0 Place…       0
#> # … with 244 more rows, and 39 more variables: TRTSDT <date>, TRTEDT <date>,
#> #   TRTDURD <dbl>, AVGDD <dbl>, CUMDOSE <dbl>, AGE <dbl>, AGEGR1 <chr>,
#> #   AGEGR1N <dbl>, AGEU <chr>, RACE <chr>, RACEN <dbl>, SEX <chr>,
#> #   ETHNIC <chr>, SAFFL <chr>, ITTFL <chr>, EFFFL <chr>, COMP8FL <chr>,
#> #   COMP16FL <chr>, COMP24FL <chr>, DISCONFL <chr>, DSRAEFL <chr>, DTHFL <chr>,
#> #   BMIBL <dbl>, BMIBLGR1 <chr>, HEIGHTBL <dbl>, WEIGHTBL <dbl>, EDUCLVL <dbl>,
#> #   DISONSDT <date>, DURDIS <dbl>, DURDSGR1 <chr>, VISIT1DT <date>, …
```

globalVariables(c(
  ".",
  ".data",
  "code",
  "decode",
  "DOMAIN",
  "IDVAR",
  "IDVARVAL",
  "QEVAL",
  "QLABEL",
  "QNAM",
  "QORIG",
  "QVAL",
  "RDOMAIN",
  "STUDYID",
  "USUBJID",
  "code_id",
  "col_name",
  "derivation",
  "derivation_id",
  "ds",
  "idvar",
  "key_seq",
  "label",
  "origin",
  "qeval",
  "supp_flag",
  "type",
  "variable",
  "merge_on"
))


#' Pipe operator
#'
#' See \code{magrittr::\link[magrittr:pipe]{\%>\%}} for details.
#'
#' @name %>%
#' @rdname pipe
#' @keywords internal
#' @export
#' @importFrom magrittr %>%
#' @usage lhs \%>\% rhs
#' @param lhs A value or the magrittr placeholder.
#' @param rhs A function call using the magrittr semantics.
#' @return The result of calling `rhs(lhs)`.
NULL

#' Apply labels to multiple variables on a data frame
#'
#' This function allows a user to apply several labels to a dataframe at once.
#'
#' @param data A data.frame or tibble
#' @param ... Named parameters in the form of variable = 'label'
#'
#' @return data with variable labels applied
#'
#' @export
#'
#' @examples
#' add_labels(
#'   mtcars,
#'   mpg = "Miles Per Gallon",
#'   cyl = "Cylinders"
#' )
#'
add_labels <- function(data, ...) {
  # Pull out ellipsis to list
  args <- list2(...)

  # Check params
  if (!inherits(data, "data.frame")) stop("Labels must be applied to a data.frame or tibble")
  if (!is_named(args)) stop("Must provide variable name and label as named arguments")

  vars_not_ex <- setdiff(names(args), names(data))
  if (length(vars_not_ex) > 0) {
    cli_abort(c(
      "x" = "The following variables do not exist in the data frame:",
      "i" = ansi_collapse(vars_not_ex, sep2 = ", ", last = ", ")
    ), call. = FALSE)
  }
  if (!all(map_lgl(args, is.character))) stop("All labels must be character")

  # Iterate the args supplied and update the variable labels in place
  walk2(names(args), args, ~ {
    attr(data[[.x]], "label") <<- .y
  })

  data
}


#' Remove labels to multiple variables on a data frame
#'
#' This function allows a user to removes all labels to a dataframe at once.
#'
#' @param data A data.frame or tibble
#'
#' @return data with variable labels applied
#' @export
#'
#' @examples
#' library(haven)
#' data <- read_xpt(metatools_example("adsl.xpt"))
#' remove_labels(data)
#'
remove_labels <- function(data) {
  # Check data
  if (!inherits(data, "data.frame")) stop("Labels must be removed from a data.frame or tibble")

  map_dfr(data, function(x) {
    attr(x, "label") <- NULL
    x
  })
}

#' Apply labels to a data frame using a metacore object
#'
#' This function leverages metadata available in a metacore object to apply
#' labels to a data frame.
#'
#' @param data A dataframe or tibble upon which labels will be applied
#' @param metacore metacore object that contains the specifications for the
#'   dataset of interest.
#' @param dataset_name `r lifecycle::badge("deprecated")` Optional string to
#'   specify the dataset that is being built. This is only needed if the metacore
#'   object provided hasn't already been subsetted.\cr
#'   Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#'   in a future release. Please use `metacore::select_dataset` to subset the
#'   `metacore` object to obtain metadata for a single dataset.
#' @param verbose Character string controlling message verbosity. One of:
#'   \describe{
#'     \item{`"message"`}{Show both warnings and messages (default)}
#'     \item{`"warn"`}{Show warnings but suppress messages}
#'     \item{`"silent"`}{Suppress all warnings and messages}
#'   }
#'
#' @return Dataframe with labels applied
#' @export
#'
#' @examples
#'
#' mc <- metacore::spec_to_metacore(
#'   metacore::metacore_example("p21_mock.xlsx"),
#'   quiet = TRUE
#' )
#' dm <- haven::read_xpt(metatools_example("dm.xpt"))
#' set_variable_labels(dm, mc, dataset_name = "DM")
set_variable_labels <- function(data, metacore, dataset_name = deprecated(),
                                verbose = c("message", "warn", "silent")) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "check_unique_keys(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }
  verify_DatasetMeta(metacore)

  verbose <- validate_verbose(verbose)

  # Grab out the var names and labels
  var_spec <- metacore$var_spec %>%
    select(variable, label)


  ns <- var_spec$variable
  labs <- var_spec$label
  dns <- names(data)

  # Are there any variables in data not in the metadata
  in_meta <- setdiff(ns, dns) # Variables in metadata not in data
  in_data <- setdiff(dns, ns) # Variables in data not in metadata

  if (length(in_meta) > 0 && check_warn(verbose)) {
    wrn <- paste0("Variables in metadata not in data:\n\t", paste0(in_meta, collapse = "\n\t"))
    warning(wrn, call. = FALSE)
  }

  if (length(in_data) > 0 && check_warn(verbose)) {
    wrn <- paste0("Variables in data not in metadata:\n\t", paste0(in_data, collapse = "\n\t"))
    warning(wrn, call. = FALSE)
  }

  # Pick out only the variables which exist in both and build list
  match <- intersect(ns, dns)
  ind <- which(ns %in% match)

  # Subset and create a named list
  ns <- ns[ind]
  labs <- labs[ind]
  names(labs) <- ns
  labs <- as.list(labs)

  # Apply the labels to the data
  args <- append(list(data), labs)
  do.call(add_labels, args)
}

# Suppress cli output during testing
options(cli.default_handler = function(...) {})

# Mock up metacore data
starwars_short <- dplyr::starwars %>% select(1:5)
var_spec <- tibble::tibble(
  variable = names(starwars_short),
  length = rep(1, 5),
  label = c("Name", "Height", "Mass", "Hair Color", "Skin Color"),
  type = c("text", "int", "int", "int", "text"),
  format = rep(NA_character_, 5),
  common = rep(FALSE, 5)
)

ds_spec <- tibble::tibble(
  dataset = "Starwars",
  structure = c(""),
  label = "Star Wars"
)

ds_vars <- tibble::tibble(
  dataset = rep("Starwars", 5),
  variable = names(starwars_short),
  order = 1:5,
  keep = rep(TRUE, 5),
  key_seq = 1:5,
  core = rep(NA_character_, 5),
  supp_flag = rep(FALSE, 5)
)

value_spec <- tibble::tibble(
  dataset = character(0),
  variable = character(0),
  type = character(0),
  origin = character(0),
  sig_dig = character(0),
  code_id = character(0),
  where = character(0),
  derivation_id = character(0)
)

derivations <- tibble::tibble(
  derivation_id = character(0),
  derivation = character(0)
)

code_id <- tibble::tibble(
  code_id = character(0),
  name = character(0),
  type = character(0),
  codes = list()
)

# This is loud and I don't want it - just need the metacore object
mc <- suppressWarnings(
  suppressMessages(
    metacore::metacore(ds_spec, ds_vars, var_spec, value_spec, derivations, code_id)
  )
)

# add_labels() tests ----
test_that("add_labels applies labels properly", {
  x <- mtcars %>%
    add_labels(
      mpg = "Miles Per Gallon",
      cyl = "Cylinders"
    )

  expect_equal(attr(x$mpg, "label"), "Miles Per Gallon")
  expect_equal(attr(x$cyl, "label"), "Cylinders")
})

test_that("add_labels errors on invalid input", {
  expect_error(add_labels(TRUE, x = "label"))
  expect_error(add_labels(mtcars, "label"))
  expect_error(add_labels(mtcars, bad = "label"))
  expect_error(add_labels(mtcars, mpg = 1))
})

# set_variable_labels() tests ----
test_that("set_variable_labels applies labels from metacore properly", {
  # Load in the metacore test object and example data
  suppressMessages(
    mc <- metacore::spec_to_metacore(metacore::metacore_example("p21_mock.xlsx"), verbose = "silent") %>%
      metacore::select_dataset("DM", verbose = "silent")
  )
  dm <- haven::read_xpt(metatools_example("dm.xpt"))

  # Set the variable labels
  dm_labeled <- set_variable_labels(dm, mc)

  # Pull out the data to check against
  labs <- purrr::map_chr(names(dm_labeled), ~ attr(dm_labeled[[.]], "label"))
  attr(labs, "label") <- "Variable Label" # This is labelled in the metacore object

  expect_equal(labs, mc$var_spec$label)
})

test_that("set_variable_labels warns on variable mismatches", {
  # This is metadata for the dplyr::starwars dataset
  mc <- suppressWarnings(
    suppressMessages(
      metacore::metacore(ds_spec, ds_vars, var_spec, value_spec, derivations, code_id)
    )
  ) %>% select_dataset("Starwars", verbose = "silent")

  # Variables in data not in metadata
  starwars_short2 <- starwars_short
  starwars_short2$new_var <- ""
  expect_warning(set_variable_labels(starwars_short2, mc))

  # Variables in metadata not in data
  mc_subset <- suppressWarnings(
    suppressMessages(
      metacore::metacore(ds_spec, ds_vars[1:4, ], var_spec[1:4, ], value_spec, derivations, code_id) %>%
        metacore::select_dataset("Starwars", verbose = "silent")
    )
  )
  expect_warning(set_variable_labels(starwars_short, mc_subset))
})

test_that("set_variable_labels respects verbose parameter", {
  load(metacore::metacore_example("pilot_SDTM.rda"))
  spec <- metacore |> select_dataset("DM", verbose = "silent")
  dm <- haven::read_xpt(metatools_example("dm.xpt"))

  # Create data with mismatch to trigger warnings
  dm_mismatch <- dm |>
    select(-RACE) |>
    mutate(EXTRAVAR = "test")

  # verbose = "silent" suppresses warnings
  expect_silent(
    set_variable_labels(dm_mismatch, spec, verbose = "silent")
  )

  # verbose = "message" shows warnings
  gotWarnings <- character(0)
  withCallingHandlers(
    {
      got <- set_variable_labels(dm_mismatch, spec, verbose = "message")
    },
    warning = function(e) {
      # Push warning onto vector
      gotWarnings <<- c(gotWarnings, conditionMessage((e)))
      invokeRestart("muffleWarning")
    }
  )

  # Ensure no unexpected warnings
  expect_equal(length(gotWarnings), 2)

  # Test that each warning I want is there
  expect_true(any(grepl("Variables in metadata not in data", gotWarnings)))
  expect_true(any(grepl("Variables in data not in metadata", gotWarnings)))

  # verbose = "warn" shows warnings
  gotWarnings <- character(0)
  withCallingHandlers(
    {
      got <- set_variable_labels(dm_mismatch, spec, verbose = "warn")
    },
    warning = function(e) {
      # Push warning onto vector
      gotWarnings <<- c(gotWarnings, conditionMessage((e)))
      invokeRestart("muffleWarning")
    }
  )

  # Ensure no unexpected warnings
  expect_equal(length(gotWarnings), 2)

  # Test that each warning I want is there
  expect_true(any(grepl("Variables in metadata not in data", gotWarnings)))
  expect_true(any(grepl("Variables in data not in metadata", gotWarnings)))

  # Invalid verbose value errors
  expect_error(
    set_variable_labels(dm, spec, verbose = "invalid"),
    "should be one of"
  )
})

# remove_labels() tests ----
test_that("remove_labels removes labels properly", {
  # Add labels first
  x <- mtcars |>
    add_labels(
      mpg = "Miles Per Gallon",
      cyl = "Cylinders"
    )

  # Verify labels exist
  expect_equal(attr(x$mpg, "label"), "Miles Per Gallon")
  expect_equal(attr(x$cyl, "label"), "Cylinders")

  # Remove labels
  x_no_labels <- remove_labels(x)

  # Verify labels are gone
  expect_null(attr(x_no_labels$mpg, "label"))
  expect_null(attr(x_no_labels$cyl, "label"))
})

test_that("remove_labels errors on invalid input", {
  expect_error(
    remove_labels("not a dataframe"),
    "Labels must be removed from a data.frame or tibble"
  )
})

#' Get path to pkg example
#'
#' pkg comes bundled with a number of sample files in its `inst/extdata`
#' directory. This function make them easy to access
#'
#' @param file Name of file. If `NULL`, the example files will be listed.
#' @export
#' @examples
#' metatools_example()
#' metatools_example("dm.xpt")
metatools_example <- function(file = NULL) {
  if (is.null(file)) {
    dir(system.file("extdata", package = "metatools"))
  } else {
    system.file("extdata", file, package = "metatools", mustWork = TRUE)
  }
}


#' Convert metacore object to just a single dataset
#'
#' @description
#' `r lifecycle::badge("deprecated")`
#' This function is deprecated as of version 0.2.0 and will be removed in a future
#' version. Dataset subsetting must now be performed via the `select_dataset`
#' function of the `metacore` package.
#'
#' @param metacore metacore object, which may or may not be subsetted
#' @param dataset_name Name of datasets which may or may not be null. If it is
#'   not null then it will be used to subset.
#'
#' @return metacore object
#' @noRd
make_lone_dataset <- function(metacore, dataset_name) {
  lifecycle::deprecate_soft(
    what = "make_lone_dataset()",
    when = "0.2.0"
  )
  if (!(nrow(metacore$ds_spec) == 1 | !is.null(dataset_name))) {
    stop("Requires either a subsetted metacore object or a dataset name", call. = FALSE)
  }
  if (!is.null(dataset_name)) {
    metacore <- select_dataset(metacore, dataset_name)
  }
  metacore
}

#' Check if messages should be displayed
#' @param verbose Verbosity level
#' @noRd
check_message <- function(verbose) {
  verbose == "message"
}

#' Check if warnings should be displayed
#' @param verbose Verbosity level
#' @noRd
check_warn <- function(verbose) {
  verbose %in% c("message", "warn")
}

#' Validate verbose parameter
#' @param verbose Verbosity level to validate
#' @noRd
validate_verbose <- function(verbose, arg = rlang::caller_arg(verbose), call = rlang::caller_env()) {
  choices <- c("message", "warn", "silent")
  tryCatch(
    match.arg(verbose, choices),
    error = function(e) {
      cli_abort(c(
        "x" = "{.arg {arg}} should be one of: {cli::ansi_collapse(choices, last = ', ')}"
      ), call = call)
    }
  )
}

# Suppress cli output during testing
options(cli.default_handler = function(...) {})

test_that("metatools_example", {
  expect_equal(metatools_example(), c("adsl.xpt", "dm.xpt"))
  expect_equal(
    file.exists(metatools_example("dm.xpt")),
    TRUE
  )

  expect_equal(
    file.exists(metatools_example("adsl.xpt")),
    TRUE
  )
})

test_that("make_lone_dataset", {
  load(metacore::metacore_example("pilot_ADaM.rda"))
  # Test deprecated function. Deprecated warning suppressed.
  suppressWarnings(expect_error(make_lone_dataset(metacore, NULL)))
})

#' @keywords internal
#' @family internal
#'
#' @importFrom cli ansi_collapse cli_abort cli_inform cli_text cli_warn qty col_red
#' @importFrom dplyr across add_count all_of anti_join any_of arrange as_tibble
#'   bind_cols bind_rows case_when distinct enquo everything filter full_join group_by
#'   group_split if_else inner_join left_join mutate pick pull rename select
#' @importFrom lifecycle deprecated deprecate_warn is_present
#' @importFrom metacore get_control_term get_keys select_dataset verify_DatasetMeta
#' @importFrom purrr discard keep map map2 map_chr map_dfr map_lgl map pmap_dfr reduce
#'   safely walk2
#' @importFrom rlang !! := as_label as_name enexpr expr is_named list2 set_names list2
#'   set_names sym is_logical
#' @importFrom stringr str_remove_all str_trim str_to_lower str_to_upper str_split
#'   str_detect str_remove str_extract str_match_all str_c
#' @importFrom tibble tibble
#' @importFrom tidyr pivot_wider unnest
#' @importFrom utils capture.output
#'
"_PACKAGE"

#' Sort Columns by Order
#'
#' This function sorts the dataset according to the order found in the
#' metacore object.
#' @param data Dataset to sort
#' @param metacore metacore object that contains the specifications for the
#'   dataset of interest.
#' @param dataset_name `r lifecycle::badge("deprecated")` Optional string to
#'   specify the dataset that is being built. This is only needed if the metacore
#'   object provided hasn't already been subsetted.\cr
#'   Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#'   in a future release. Please use `metacore::select_dataset` to subset the
#'   `metacore` object to obtain metadata for a single dataset.
#'
#' @return dataset with ordered columns
#' @export
#'
#' @examples
#' library(metacore)
#' library(haven)
#' library(magrittr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' data <- read_xpt(metatools_example("adsl.xpt"))
#' order_cols(data, spec)
order_cols <- function(data, metacore, dataset_name = deprecated()) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "order_cols(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }
  verify_DatasetMeta(metacore)
  var_ord <- metacore$ds_vars %>%
    filter(!is.na(order)) %>%
    arrange(order) %>%
    pull(variable) %>%
    keep(~ . %in% names(data))

  data %>%
    select(all_of(var_ord), everything())
}


#' Sort Rows by Key Sequence
#'
#' This function sorts the dataset according to the key sequence found in the
#' metacore object.
#' @param data Dataset to sort
#' @param metacore metacore object that contains the specifications for the
#'   dataset of interest.
#' @param dataset_name `r lifecycle::badge("deprecated")` Optional string to
#'   specify the dataset that is being built. This is only needed if the metacore
#'   object provided hasn't already been subsetted.\cr
#'   Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#'   in a future release. Please use `metacore::select_dataset` to subset the
#'   `metacore` object to obtain metadata for a single dataset.
#'
#' @return dataset with ordered columns
#' @export
#'
#' @examples
#' library(metacore)
#' library(haven)
#' library(magrittr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' data <- read_xpt(metatools_example("adsl.xpt"))
#' sort_by_key(data, spec)
sort_by_key <- function(data, metacore, dataset_name = deprecated()) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "sort_by_key(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }
  verify_DatasetMeta(metacore)
  var_ord <- metacore$ds_vars %>%
    filter(!is.na(key_seq)) %>%
    arrange(key_seq) %>%
    pull(variable)

  data %>%
    arrange(across(all_of(var_ord)))
}

# order_cols() tests ----
test_that("order_cols reorders columns to match metacore spec", {
  load(metacore::metacore_example("pilot_ADaM.rda"))
  spec <- metacore %>% select_dataset("ADSL", verbose = "silent")
  data <- haven::read_xpt(metatools_example("adsl.xpt"))

  result <- data %>%
    select(AGE, SITEID, everything()) %>%
    order_cols(spec)

  expect_equal(result, data)
})

test_that("order_cols handles extra columns not in spec", {
  load(metacore::metacore_example("pilot_ADaM.rda"))
  spec <- metacore %>% select_dataset("ADSL", verbose = "silent")
  data <- haven::read_xpt(metatools_example("adsl.xpt"))

  result <- data %>%
    select(AGE, everything(), -SITEID) %>%
    order_cols(spec)

  expected <- select(data, -SITEID)
  expect_equal(result, expected)
})

test_that("order_cols handles additional columns beyond spec", {
  load(metacore::metacore_example("pilot_ADaM.rda"))
  spec <- metacore %>% select_dataset("ADSL", verbose = "silent")
  data <- haven::read_xpt(metatools_example("adsl.xpt"))

  result <- data %>%
    select(AGE, SITEID, everything()) %>%
    mutate(foo = "game") %>%
    order_cols(spec)

  expected <- mutate(data, foo = "game")
  expect_equal(result, expected)
})

# sort_by_key() tests ----
test_that("sort_by_key sorts data by key variables from metacore spec", {
  load(metacore::metacore_example("pilot_ADaM.rda"))
  spec <- metacore %>% select_dataset("ADSL", verbose = "silent")
  data <- haven::read_xpt(metatools_example("adsl.xpt"))

  result <- data %>%
    arrange(TRT01P, AGE) %>%
    sort_by_key(spec)

  expect_equal(result, data)
})

#' Dash to Equation
#'
#' Converts strings that are #-# style to a logical expression (but in a string format)
#' @param string
#'
#' @return string
#' @noRd
dash_to_eq <- function(string) {
  front <- str_extract(string, "^.*(?=\\-)")
  front_eq <- if_else(str_detect(front, "<|>|="), front, paste0(">=", front))
  back <- str_extract(string, "(?<=\\-).*$")
  back_eq <- if_else(str_detect(back, "<|>|="), back, paste0("<=", back))
  paste0("x", front_eq, " & x", back_eq)
}


#' Create Subgroups
#'
#' @param ref_vec Vector of numeric values
#' @param grp_defs Vector of strings with groupings defined. Format must be
#'   either: <00, >=00, 00-00, or  00-<00
#' @param grp_labs Vector of strings with labels defined. The labels correspond
#'   to the associated `grp_defs`. i.e., "12-17" may translate to "12-17 years".
#'   If no `grp_labs` specified then `grp_defs` will be used.
#'
#' @return Character vector of the values in the subgroups
#' @export
#'
#' @examples
#' create_subgrps(c(1:10), c("<2", "2-5", ">5"))
#' create_subgrps(c(1:10), c("<=2", ">2-5", ">5"))
#' create_subgrps(c(1:10), c("<2", "2-<5", ">=5"))
#' create_subgrps(c(1:10), c("<2", "2-<5", ">=5"), c("<2 years", "2-5 years", ">=5 years"))
create_subgrps <- function(ref_vec, grp_defs, grp_labs = NULL) {
  if (!is.numeric(ref_vec)) {
    cli_abort("ref_vec must be numeric")
  }
  if (is.null(grp_labs)) {
    grp_labs <- grp_defs
  }

  # Create equations used to derive the subgroups
  equations <- case_when(
    str_detect(grp_defs, "-") ~ paste0("function(x){if_else(", dash_to_eq(grp_defs), ", '", grp_labs, "', '')}"),
    str_detect(grp_defs, "^(<\\s?=|>\\s?=|<|>)\\s?\\d+") ~ paste0("function(x){if_else(x", grp_defs, ",'", grp_labs, "', '')}"),
    TRUE ~ NA_character_
  )

  # Apply equations
  if (all(!is.na(equations))) {
    functions <- equations %>%
      map(~ eval(parse(text = .)))
    out <- functions %>%
      map(~ .(ref_vec)) %>%
      reduce(str_c) %>%
      replace(. == "", NA)
  } else {
    na_index <- which(is.na(equations))
    bad_defs <- grp_defs[na_index]
    cli_abort(paste(
      "Unable to decipher the following group definition{?s}: {bad_defs}.",
      "Please check your controlled terminology."
    ))
  }
  # Find non-exclusive subgroups i.e., values that have been mapped to two groups
  non_excl <- out |>
    discard(is.na) |>
    map(~ grp_labs[str_detect(.x, grp_labs)]) |>
    keep(~ length(.) > 1) |>
    unique()

  # Throw error if groups are not exclusive
  if (length(non_excl) > 0) {
    msg <- map_chr(non_excl, ~ {
      items <- paste(.x, collapse = ", ")
    }) %>%
      paste0(seq_along(.), ". ", .)

    cli_abort(c(
      "Group definitions are not exclusive. Please check your controlled terminology",
      "The following group definitions overlap:",
      msg
    ))
  }
  out
}


#' Create Variable from Codelist
#'
#' This functions uses code/decode pairs from a metacore object to create new
#' variables in the data
#'
#' @param data Dataset that contains the input variable
#' @param metacore A metacore object to get the codelist from. This should be a
#'   subsetted metacore object (of subclass `DatasetMeta`) created using
#'   `metacore::select_dataset`.
#' @param input_var Name of the variable that will be translated for the new
#'   column
#' @param out_var Name of the output variable. Note: Unless a codelist is provided
#'   the grouping will always be from the code of the codelist associates with
#'   `out_var`.
#' @param codelist Optional argument to supply a codelist. Must be a data.frame
#'   with `code` and `decode` columns such as those created by the function
#'   `metacore::get_control_term`. If no codelist is provided the codelist
#'   associated with the column supplied to `out_var` will be used. By default
#'   `codelist` is `NULL`.
#' @param decode_to_code Direction of the translation. Default value is `TRUE`,
#'    i.e., assumes the  `input_var` is the decode column of the codelist.
#'    Set to `FALSE` if the `input_var` is the code column of the codelist.
#' @param strict A logical value indicating whether to perform strict checking
#'   against the codelist. If `TRUE` will issue a warning if values in the `input_var`
#'   column are not present in the codelist. If `FALSE` no warning is issued and
#'   values not present in the codelist will likely result in `NA` results.
#'
#' @return Dataset with a new column added
#' @export
#'
#' @examples
#' library(metacore)
#' library(tibble)
#' data <- tribble(
#'   ~USUBJID, ~VAR1, ~VAR2,
#'   1, "M", "Male",
#'   2, "F", "Female",
#'   3, "F", "Female",
#'   4, "U", "Unknown",
#'   5, "M", "Male",
#' )
#' spec <- spec_to_metacore(metacore_example("p21_mock.xlsx"), quiet = TRUE)
#' dm_spec <- select_dataset(spec, "DM", quiet = TRUE)
#' create_var_from_codelist(data, dm_spec, VAR2, SEX)
#' create_var_from_codelist(data, dm_spec, "VAR2", "SEX")
#' create_var_from_codelist(data, dm_spec, VAR1, SEX, decode_to_code = FALSE)
#'
#' # Example providing a custom codelist
#' # This example also reverses the direction of translation
#' load(metacore_example("pilot_ADaM.rda"))
#' adlb_spec <- select_dataset(metacore, "ADLBC", quiet = TRUE)
#' adlb <- tibble(PARAMCD = c("ALB", "ALP", "ALT", "AST", "BILI", "BUN"))
#' create_var_from_codelist(
#'   adlb,
#'   adlb_spec,
#'   PARAMCD,
#'   PARAM,
#'   codelist = get_control_term(adlb_spec, PARAMCD),
#'   decode_to_code = FALSE,
#'   strict = FALSE
#' )
#'
#' \dontrun{
#' # Example expecting warning where `strict` == `TRUE`
#' adlb <- tibble(PARAMCD = c("ALB", "ALP", "ALT", "AST", "BILI", "BUN", "DUMMY1", "DUMMY2"))
#' create_var_from_codelist(
#'   adlb,
#'   adlb_spec,
#'   PARAMCD,
#'   PARAM,
#'   codelist = get_control_term(adlb_spec, PARAMCD),
#'   decode_to_code = FALSE,
#'   strict = TRUE
#' )
#' }
create_var_from_codelist <- function(data, metacore, input_var, out_var, codelist = NULL,
                                     decode_to_code = TRUE, strict = TRUE) {
  verify_DatasetMeta(metacore)

  # Use codelist if provided, else use codelist of the out_var
  if (!missing(codelist)) {
    code_translation <- codelist
  } else {
    code_translation <- get_control_term(metacore, {{ out_var }})
  }

  if (is.vector(code_translation) | !("decode" %in% names(code_translation))) {
    cli_abort("Expecting 'code_decode' type of control terminology. Actual \\
type is {typeof(code_translation)}. Check the structure of the codelist in the \\
{.obj metacore} object using {.fn View}.")
  }

  # Check decode_to_code is logical and set direction of translation
  if (!is_logical(decode_to_code)) {
    cli_abort("{.arg decode_to_code} must be either TRUE or FALSE.")
  }

  ref_var <- if (decode_to_code) "decode" else "code"
  new_var <- if (decode_to_code) "code" else "decode"

  # Pull data values and codelist values to check inconsistent overlap
  values <- data |> pull({{ input_var }})
  codelist <- code_translation |> pull(ref_var)

  miss <- setdiff(values, codelist)
  n_miss <- length(miss)
  if (strict == TRUE && n_miss > 0) {
    cli_warn(
      "In {.fn create_var_from_codelist}: The following {qty(n_miss)}value{?s}
present in the input dataset {qty(n_miss)}{?is/are} not present in the codelist: {miss}"
    )
  }

  input_var_str <- as_label(enexpr(input_var)) |>
    str_remove_all("\"")

  # Coerce join column to character to ensure join if input var is numeric
  data <- data |> mutate(merge_on := as.character(.data[[input_var_str]]))
  code_translation <- code_translation |>
    mutate(
      decode = as.character(decode),
      code = as.character(code)
    )

  out <- data |>
    left_join(code_translation, by = set_names(ref_var, "merge_on")) |>
    rename({{ out_var }} := !!sym(new_var)) |>
    select(-merge_on)

  # Optionally coerce to numeric if the output values are numeric
  if (all(str_detect(code_translation[[new_var]], "^-?\\d*$"))) {
    out <- out |>
      mutate({{ out_var }} := as.numeric({{ out_var }}))
  }

  out
}


#' Create Categorical Variable from Codelist
#'
#' Using the grouping from either the `decode_var` or `code_var` and a reference
#' variable (`ref_var`) it will create a categorical variable and the numeric
#' version of that categorical variable.
#'
#' @param data Dataset with reference variable in it
#' @param metacore A metacore object to get the codelist from. If the
#'   variable has different codelists for different datasets the metacore object
#'   will need to be subsetted using `select_dataset` from the metacore package.
#' @param ref_var Name of variable to be used as the reference i.e AGE when
#'   creating AGEGR1
#' @param grp_var Name of the new grouped variable
#' @param num_grp_var Name of the new numeric decode for the grouped variable.
#'   This is optional if no value given no variable will be created
#' @param create_from_decode Sets the `decode` column of the codelist as the column
#'   from which the variable will be created. By default the column is `code`.
#' @param strict A logical value indicating whether to perform strict checking
#'   against the codelist. If `TRUE` will issue a warning if values in the `ref_var`
#'   column do not fit into the group definitions for the codelist in `grp_var`.
#'   If `FALSE` no warning is issued and values not defined by the codelist will
#'   likely result in `NA` results.
#'
#' @return dataset with new column added
#' @export
#' @examples
#' library(metacore)
#' library(haven)
#' library(dplyr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' dm <- read_xpt(metatools_example("dm.xpt")) %>%
#'   select(USUBJID, AGE)
#' # Grouping Column Only
#' create_cat_var(dm, spec, AGE, AGEGR1)
#' # Grouping Column and Numeric Decode
#' create_cat_var(dm, spec, AGE, AGEGR1, AGEGR1N)
create_cat_var <- function(data, metacore, ref_var, grp_var, num_grp_var = NULL,
                           create_from_decode = FALSE, strict = TRUE) {
  verify_DatasetMeta(metacore)
  ct <- get_control_term(metacore, {{ grp_var }})
  if (is.vector(ct) | !("decode" %in% names(ct))) {
    cli_abort("Expecting 'code_decode' type of control terminology. Please check metacore object")
  }

  # Assign group definitions and labels
  grp_defs <- pull(ct, code)
  grp_labs <- if (create_from_decode) pull(ct, decode) else grp_defs

  out <- data %>%
    mutate({{ grp_var }} := create_subgrps({{ ref_var }}, grp_defs, grp_labs))

  if (!is.null(enexpr(num_grp_var))) {
    out <- out %>%
      create_var_from_codelist(metacore, {{ grp_var }}, {{ num_grp_var }})
  }

  missing <- out |>
    pull({{ grp_var }}) |>
    is.na() |>
    which() |>
    length()
  if (strict && missing > 0) {
    cli_warn(paste(
      "There {qty(missing)} {?is/are} {missing} {qty(missing)} observation{?s}",
      "in {as_name(enquo(ref_var))} that {qty(missing)} {?does/do} not fit into",
      "the provided categories for {as_name(enquo(grp_var))}. Please check your",
      "controlled terminology."
    ))
  }
  out
}


#' Convert Variable to Factor with Levels Set by Control Terms
#'
#' This functions takes a dataset, a metacore object and a variable name. Then
#' looks at the metacore object for the control terms for the given variable and
#' uses that to convert the variable to a factor with those levels. If the
#' control terminology is a code list, the code column will be used. The
#' function fails if the control terminology is an external library
#' @param data A dataset containing the variable to be modified
#' @param metacore A metacore object to get the codelist from. If the
#'   variable has different codelists for different datasets the metacore object
#'   will need to be subsetted using `select_dataset` from the metacore package
#' @param var Name of variable to change
#'
#' @return Dataset with variable changed to a factor
#' @export
#'
#' @examples
#' library(metacore)
#' library(haven)
#' library(dplyr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' dm <- read_xpt(metatools_example("dm.xpt")) %>%
#'   select(USUBJID, SEX, ARM)
#' # Variable with codelist control terms
#' convert_var_to_fct(dm, spec, SEX)
#' # Variable with permitted value control terms
#' convert_var_to_fct(dm, spec, ARM)
convert_var_to_fct <- function(data, metacore, var) {
  verify_DatasetMeta(metacore)
  code_translation <- get_control_term(metacore, {{ var }})
  var_str <- as_label(enexpr(var)) %>%
    str_remove_all("\"")
  if (is.vector(code_translation)) {
    levels <- code_translation
  } else if ("code" %in% names(code_translation)) {
    levels <- code_translation$code
  } else {
    stop("We currently don't have the ability to use external libraries")
  }
  if (!var_str %in% names(data)) {
    stop(paste(var_str, "cannot be found in the dataset. Please create variable before converting to factor"))
  }
  data %>%
    mutate({{ var }} := factor({{ var }}, levels = levels))
}

# Suppress cli output during testing
options(cli.default_handler = function(...) {})

spec <- metacore::spec_to_metacore(metacore::metacore_example("p21_mock.xlsx"), verbose = "silent")
dm_spec <- select_dataset(spec, "DM", verbose = "silent")
load(metacore::metacore_example("pilot_ADaM.rda"))
adsl_spec <- metacore %>% select_dataset("ADSL", verbose = "silent")
dm <- haven::read_xpt(metatools_example("dm.xpt"))


test_that("create_subgrps", {
  expect_equal(
    create_subgrps(c(1:10), c("<2", "2-5", ">5")),
    c("<2", "2-5", "2-5", "2-5", "2-5", ">5", ">5", ">5", ">5", ">5")
  )

  expect_equal(
    create_subgrps(c(1:10), c("<=2", ">2-5", ">5")),
    c("<=2", "<=2", ">2-5", ">2-5", ">2-5", ">5", ">5", ">5", ">5", ">5")
  )

  expect_equal(
    create_subgrps(c(1:10), c("<2", "2-<5", ">=5")),
    c("<2", "2-<5", "2-<5", "2-<5", ">=5", ">=5", ">=5", ">=5", ">=5", ">=5")
  )


  expect_equal(
    create_subgrps(c(1:10, NA), c("<2", "2-5", ">5")),
    c("<2", "2-5", "2-5", "2-5", "2-5", ">5", ">5", ">5", ">5", ">5", NA)
  )
})

test_that("create_var_from_codelist", {
  data <- tibble::tribble(
    ~USUBJID, ~VAR1, ~VAR2,
    1, "M", "Male",
    2, "F", "Female",
    3, "F", "Female",
    4, "U", "Unknown",
    5, "M", "Male",
  )

  manual_data <- data %>%
    mutate(SEX = VAR1)
  expect_equal(create_var_from_codelist(data, dm_spec, VAR2, SEX), manual_data)
  expect_equal(create_var_from_codelist(data, dm_spec, "VAR2", "SEX"), manual_data)
  manual_data2 <- data %>%
    mutate(SEX = VAR2)
  expect_equal(
    create_var_from_codelist(data, dm_spec, VAR1, SEX, decode_to_code = FALSE),
    manual_data2
  )
  # Test numeric
  num_out <- dm %>%
    filter(ARMCD != "Scrnfail") %>%
    mutate(TRT01P = ARM) %>%
    select(TRT01P, ARMCD) %>%
    create_var_from_codelist(adsl_spec, TRT01P, TRT01PN) %>%
    head() %>%
    pull(TRT01PN)
  expect_equal(num_out, c(0, 0, 81, 54, 81, 0))

  # Test provide custom codelist
  load(metacore::metacore_example("pilot_ADaM.rda"))
  adlb_spec <- metacore::select_dataset(metacore, "ADLBC", verbose = "silent")
  data <- tibble::tibble(
    PARAMCD = c("ALB", "ALP", "ALT", "DUMMY", "DUMMY2")
  )
  compare <- tibble::tibble(
    PARAMCD = c("ALB", "ALP", "ALT", "DUMMY", "DUMMY2"),
    PARAM = c("Albumin (g/L)", "Alkaline Phosphatase (U/L)", "Alanine Aminotransferase (U/L)", NA, NA)
  )

  create_var_from_codelist(
    data = data,
    metacore = adlb_spec,
    input_var = PARAMCD,
    out_var = PARAM,
    codelist = get_control_term(adlb_spec, PARAMCD),
    decode_to_code = FALSE,
    strict = FALSE
  ) |>
    select(PARAMCD, PARAM) |>
    expect_equal(compare)

  # Test character variable warning where arg `strict == TRUE` / single issue case
  create_var_from_codelist(
    data = data,
    metacore = adlb_spec,
    input_var = PARAMCD,
    out_var = PARAM,
    codelist = metacore::get_control_term(adlb_spec, PARAMCD),
    decode_to_code = FALSE,
    strict = TRUE
  ) |>
    expect_warning()

  # Test character variable warning where arg `strict == TRUE` / multiple issue case
  data <- tibble::tibble(
    PARAMCD = c("ALB", "ALP", "ALT", "DUMMY", "DUMMY2")
  )

  create_var_from_codelist(
    data = data,
    metacore = adlb_spec,
    input_var = PARAMCD,
    out_var = PARAM,
    codelist = get_control_term(adlb_spec, PARAMCD),
    decode_to_code = FALSE,
    strict = TRUE
  ) |>
    expect_warning()

  # Test numeric variable used as input_var (strict == FALSE)
  data2 <- tibble::tibble(
    PARAMN = c(18, 19, 20, 99)
  )
  compare2 <- tibble::tibble(
    PARAMN = c(18, 19, 20, 99),
    PARAM = c("Sodium (mmol/L)", "Potassium (mmol/L)", "Chloride (mmol/L)", NA)
  )

  create_var_from_codelist(
    data = data2,
    metacore = adlb_spec,
    input_var = PARAMN,
    out_var = PARAM,
    codelist = get_control_term(adlb_spec, PARAMN),
    decode_to_code = FALSE,
    strict = FALSE
  ) |>
    select(PARAMN, PARAM) |>
    expect_equal(compare2)

  # Test numeric variable used as input_var / single issue case (strict == TRUE)
  create_var_from_codelist(
    data = data2,
    metacore = adlb_spec,
    input_var = PARAMN,
    out_var = PARAM,
    codelist = get_control_term(adlb_spec, PARAMN),
    decode_to_code = FALSE,
    strict = TRUE
  ) |>
    expect_warning()

  # Test numeric variable used as input_var / multiple issue case (strict == TRUE)
  data3 <- tibble::tibble(
    PARAMN = c(18, 19, 20, 99, 999)
  )

  create_var_from_codelist(
    data = data3,
    metacore = adlb_spec,
    input_var = PARAMN,
    out_var = PARAM,
    codelist = get_control_term(adlb_spec, PARAMN),
    decode_to_code = FALSE,
    strict = TRUE
  ) |>
    expect_warning()

  # Test numeric codelist with negative numbers
  data <- tibble::tibble(
    USUBJID = c(1, 2, 3, 4, 5),
    VAR = c("Male", "Female", "Female", "Unknown", "Male")
  )

  manual_data <- tibble::tibble(
    USUBJID = c(1, 2, 3, 4, 5),
    VAR = c("Male", "Female", "Female", "Unknown", "Male"),
    SEX = c(2, 1, 1, -999, 2),
  )

  dm_spec <- metacore::metacore(
    verbose = "silent",
    ds_spec = spec$ds_spec,
    ds_vars = spec$ds_vars,
    var_spec = spec$var_spec,
    value_spec = mutate(spec$value_spec, code_id = case_when(code_id != "SEX" ~ NA, TRUE ~ "SEX")),
    derivations = spec$derivations,
    supp = spec$supp,
    codelist = structure(
      list(
        code_id = structure("SEX", label = "ID of the Code List"),
        name = "SEX",
        type = structure("code_decode", label = "Code List/Permitted Values/External Library"),
        codes = structure(list(structure(list(code = c(1, 2, -999), decode = c("Female", "Male", "Unknown")), row.names = c(NA, -3L), class = c("tbl_df", "tbl", "data.frame"))), label = "List of Codes")
      ),
      row.names = c(NA, -1L), class = c("tbl_df", "tbl", "data.frame")
    )
  ) %>%
    select_dataset("DM", verbose = "silent")

  create_var_from_codelist(data, dm_spec, VAR, SEX) %>%
    expect_equal(manual_data)

  # Test for Variable not in specs
  expect_error(
    create_var_from_codelist(data, adlb_spec, VAR2, FOO),
    regexp = cli_inform(c("!" = "FOO not found in the value_spec table. Please check the variable name"))
  )
})

test_that("create_cat_var", {
  # Create manual dataset to check against
  man_dat <- tibble::tribble(
    ~AGEGR1,     ~n,
    "65-80",    172,
    "<65",       42,
    ">80",       92,
  )

  man_dat_labs <- tibble::tribble(
    ~AGEGR2,        ~n,
    "18-64 years",  42,
    "65-80 years",  172,
    ">80 years",    92,
  )
  # Grouping col only
  auto_dat <- create_cat_var(dm, adsl_spec, AGE, AGEGR1) %>%
    group_by(AGEGR1) %>%
    dplyr::summarise(n = dplyr::n())
  expect_equal(auto_dat, man_dat)
  # Grouping Column and Numeric Decode
  grp_num_dat <- create_cat_var(dm, adsl_spec, AGE, AGEGR1, AGEGR1N)
  grp_num_dat %>%
    group_by(AGEGR1) %>%
    dplyr::summarise(n = dplyr::n()) %>%
    expect_equal(auto_dat)
  grp_num_dat %>%
    pull(AGEGR1N) %>%
    unique() %>%
    expect_equal(c(1:3))

  # Grouping column and numeric decode, build from decode == TRUE
  decode_num_dat <- create_cat_var(dm, adsl_spec, AGE, AGEGR2, AGEGR2N, TRUE)
  decode_num_dat %>%
    group_by(AGEGR2) %>%
    dplyr::summarise(n = dplyr::n()) %>%
    expect_equal(man_dat_labs)
  decode_num_dat %>%
    pull(AGEGR2N) %>%
    unique() %>%
    expect_equal(c(1:3))

  # Test error 'unable to decipher group definition'
  bad_ct <- adsl_spec$codelist |>
    filter(name == "AGEGR1") |>
    pull(codes) |>
    purrr::pluck(1) |>
    tibble::add_row(code = "DUMMY", decode = "DUMMY")

  codelist <- adsl_spec$codelist |> filter(name == "AGEGR1")
  codelist$codes <- list(bad_ct)

  spec2 <- suppressWarnings(metacore::metacore(
    adsl_spec$ds_spec,
    adsl_spec$ds_vars,
    adsl_spec$var_spec,
    adsl_spec$value_spec,
    adsl_spec$derivations,
    codelist = codelist,
    supp = adsl_spec$supp
  )) %>%
    select_dataset("ADSL", verbose = "silent")

  create_cat_var(dm, spec2, AGE, AGEGR1, AGEGR1N, TRUE) |>
    expect_error("Unable to decipher the following group definition: DUMMY. Please check your controlled terminology.")

  # Test error 'group definitions are not exclusive'
  bad_ct <- adsl_spec$codelist |>
    filter(name == "AGEGR1") |>
    pull(codes) |>
    purrr::pluck(1) |>
    tibble::add_row(code = "18-64", decode = "18-64 years")

  codelist <- adsl_spec$codelist |> filter(name == "AGEGR1")
  codelist$codes <- list(bad_ct)

  spec2 <- suppressWarnings(metacore::metacore(
    adsl_spec$ds_spec,
    adsl_spec$ds_vars,
    adsl_spec$var_spec,
    adsl_spec$value_spec,
    adsl_spec$derivations,
    codelist = codelist,
    supp = adsl_spec$supp
  )) %>%
    select_dataset("ADSL", verbose = "silent")

  create_cat_var(dm, spec2, AGE, AGEGR1, AGEGR1N, create_from_decode = TRUE) |>
    expect_error("Group definitions are not exclusive. Please check your controlled terminology")

  # Test error 'value exists that is not defined in controlled terminology
  dm2 <- dm |>
    tibble::add_row(AGE = 15) |>
    tibble::add_row(AGE = 16)
  x <- create_cat_var(dm2, adsl_spec, AGE, AGEGR2, create_from_decode = TRUE) |>
    expect_warning()

  # Test errors
  expect_error(create_cat_var(dm, spec, AGE, ARM))
})

test_that("convert_var_to_fct", {
  # Codelist variable
  convert_var_to_fct(dm, dm_spec, SEX) %>%
    pull(SEX) %>%
    levels() %>%
    expect_equal(c("F", "M", "U"))
  # Param list variable
  convert_var_to_fct(dm, dm_spec, ARM) %>%
    pull(ARM) %>%
    levels() %>%
    expect_equal(c("Screen Failure", "Placebo", "Xanomeline Low Dose", "Xanomeline High Dose"))
  # Errors

  # Note although AELLT isn't in DM it will fail because it is an
  # external lib before it fails for not being the dataset
  expect_error(convert_var_to_fct(dm, dm_spec, AELLT))
  expect_error(convert_var_to_fct(dm, dm_spec, FOO))
})

#' Build a dataset from derived
#'
#' This function builds a dataset out of the columns that just need to be pulled
#' through. So any variable that has a derivation in the format of
#' 'dataset.variable' will be pulled through to create the new dataset. When
#' there are multiple datasets present, they will be joined by the shared
#' `key_seq` variables. These columns are often called 'Predecessors' in ADaM,
#' but this is not universal so that is optional to specify.
#' @param metacore metacore object that contains the specifications for the
#'   dataset of interest.
#' @param ds_list Named list of datasets that are needed to build the from. If
#'   the list is unnamed,then it will use the names of the objects.
#' @param dataset_name `r lifecycle::badge("deprecated")` Optional string to
#'   specify the dataset that is being built. This is only needed if the metacore
#'   object provided hasn't already been subsetted.\cr
#'   Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#'   in a future release. Please use `metacore::select_dataset` to subset the
#'   `metacore` object to obtain metadata for a single dataset.
#' @param predecessor_only By default `TRUE`, so only variables with the origin
#'   of 'Predecessor' will be used. If `FALSE` any derivation matching the
#'   dataset.variable will be used.
#' @param keep String to determine which columns from the original datasets
#'   should be kept
#'   - "FALSE" (default):  only columns that are also present in the ADaM
#'                            specification are kept in the output.
#'   - "ALL":              all original columns are carried through to the
#'                            ADaM, including those that have been renamed.
#'                            e.g. if DM.ARM is a predecessor to DM.TRT01P,
#'                            both ARM and TRT01P will be present as columns
#'                            in the ADaM output.
#'   - "PREREQUISITE":     columns are retained if they are required for future
#'                            derivations in the specification. Additional
#'                            prerequisite columns are identified as columns
#'                            that appear in the 'derivation' column of the
#'                            metacore object in the format "DATASET.VARIABLE",
#'                            but not as direct predecessors. Predecessors are
#'                            defined as columns where the derivation is a 1:1
#'                            copy of a column in a source dataset.
#'                            e.g. derivation = "VS.VSTESTCD" is a predecessor,
#'                            while derivation = "Value of VS.VSSTRESN where
#'                            VS.VSTESTCD == 'Heart Rate'" contains both
#'                            VS.VSTESTCD and VS.VSSTRESN as prerequisites, and
#'                            these columns will be kept through to the ADaM.
#' @param verbose Character string controlling message verbosity. One of:
#'   \describe{
#'     \item{`"message"`}{Show both warnings and messages (default)}
#'     \item{`"warn"`}{Show warnings but suppress messages}
#'     \item{`"silent"`}{Suppress all warnings and messages}
#'   }
#'
#' @return dataset
#' @export
#'
#' @examples
#' library(metacore)
#' library(haven)
#' library(magrittr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' ds_list <- list(DM = read_xpt(metatools_example("dm.xpt")))
#' build_from_derived(spec, ds_list, predecessor_only = FALSE)
build_from_derived <- function(metacore, ds_list, dataset_name = deprecated(),
                               predecessor_only = TRUE, keep = FALSE,
                               verbose = c("message", "warn", "silent")) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "build_from_derived(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }
  verify_DatasetMeta(metacore)

  verbose <- validate_verbose(verbose)

  # Deprecate KEEP = TRUE
  keep <- match.arg(as.character(keep), c("TRUE", "FALSE", "ALL", "PREREQUISITE"))
  if (keep == "TRUE") {
    cli_warn(paste0(
      "Setting 'keep' = TRUE has been superseded",
      ", and will be unavailable in future releases. Please consider setting ",
      "'keep' equal to 'ALL' or 'PREREQUISITE'."
    ))
  }

  derirvations <- metacore$derivations %>%
    mutate(derivation = trimws(derivation))

  if (predecessor_only) {
    limited_dev_ids <- metacore$value_spec %>%
      filter(str_detect(str_to_lower(origin), "predecessor")) %>%
      pull(derivation_id)

    derirvations <- derirvations %>%
      filter(derivation_id %in% limited_dev_ids)
    if (nrow(derirvations) == 0) {
      stop("No predecessor variables found please check your metacore object")
    }
  }

  vars_to_pull_through <- derirvations %>%
    filter(str_detect(derivation, "^\\w*\\.[a-zA-Z0-9]*$"))

  # To lower so it is flexible about how people name their ds list
  vars_w_ds <- vars_to_pull_through %>%
    mutate(ds = str_extract(derivation, "^\\w*(?=\\.)") %>%
      str_to_lower())
  ds_names <- vars_w_ds %>%
    pull(ds) %>%
    unique()
  if (is.null(names(ds_list))) {
    names(ds_list) <- deparse(substitute(ds_list)) %>%
      str_remove("list\\s?\\(") %>%
      str_remove("\\)s?$") %>%
      str_split(",\\s?") %>%
      unlist()
  }
  names(ds_list) <- names(ds_list) %>%
    str_to_lower()
  if (!all(ds_names %in% names(ds_list))) {
    unknown <- keep(names(ds_list), ~ !. %in% ds_names)
    if (length(unknown) > 0 && check_warn(verbose)) {
      warning(paste0("The following dataset(s) have no predecessors and will be ignored:\n"),
        paste0(unknown, collapse = ", "),
        call. = FALSE
      )
    }
    ds_using <- discard(names(ds_list), ~ . %in% unknown) %>%
      str_to_upper() %>%
      paste0(collapse = ", ")

    if (check_message(verbose)) {
      message(paste0(
        "Not all datasets provided. Only variables from ",
        ds_using,
        " will be gathered."
      ))
    }

    # Filter out any variable that come from datasets that aren't present
    vars_w_ds <- vars_w_ds %>%
      filter(ds %in% names(ds_list))
  }

  ds_keys <- metacore$ds_vars %>%
    filter(!is.na(key_seq)) %>%
    pull(variable)

  joining_vals_to_add <- ds_list %>%
    map(function(x) {
      names(x) %>%
        keep(~ . %in% ds_keys)
    })

  join_by <- joining_vals_to_add %>%
    reduce(intersect)
  additional_vals <- tibble(
    ds = names(ds_list),
    variable = joining_vals_to_add
  ) %>%
    unnest(variable) %>%
    mutate(col_name = variable)

  vars_w_ds %>%
    mutate(col_name = str_extract(derivation, "(?<=\\.).*")) %>%
    inner_join(metacore$value_spec, ., by = "derivation_id") %>%
    select(variable, ds, col_name) %>%
    bind_rows(additional_vals) %>%
    group_by(ds) %>%
    group_split() %>%
    map(get_variables, ds_list, keep, derirvations) %>%
    prepare_join(join_by, names(ds_list), verbose) %>%
    reduce(full_join, by = join_by)
}

#' Internal functions to get variables from a dataset list
#'
#' This function is used with `build_from_derived` to build a dataset of columns
#' that are pulled directly from other datasets
#'
#' @param x Dataset with the old and new variable name and dataset name
#' @param ds_list List of datasets
#' @param keep boolean if old columns should be kept
#'
#' @return datasets
#' @noRd
get_variables <- function(x, ds_list, keep, derivations) {
  ds_name <- unique(x$ds)
  data <- ds_list[[ds_name]]
  rename_vec <- set_names(x$col_name, x$variable)
  if (keep == "TRUE") {
    # Don't drop predecessor columns
    out <- data %>%
      select(x$col_name) %>%
      mutate(across(all_of(rename_vec)))
  } else if (keep == "FALSE") {
    # Drop predecessor columns
    out <- data %>%
      select(x$col_name) %>%
      mutate(across(all_of(rename_vec))) %>%
      select(x$variable)
  } else if (keep == "ALL") {
    # Keep all cols from original datasets
    out <- data %>%
      mutate(across(all_of(rename_vec)))
  } else if (keep == "PREREQUISITE") {
    # Keep all columns required for future derivations
    # Find all "XX.XXXXX"
    future_derivations <- derivations %>%
      select(derivation) %>%
      filter(!str_detect(derivation, "^[A-Z0-9a-z]+\\.[A-Z0-9a-z]+$"))

    prereq_vector <- str_match_all(future_derivations$derivation, "([A-Z0-9a-z]+)\\.([A-Z0-9a-z]+)")

    # Bind into matrix + remove dups
    prereq_matrix <- do.call(rbind, prereq_vector) %>%
      unique()

    # Subset to those present in current dataset
    prereq_cols <- subset(prereq_matrix, tolower(prereq_matrix[, 2]) == tolower(ds_name))[, 3]

    out <- data %>%
      select(c(x$col_name, all_of(prereq_cols))) %>%
      mutate(across(all_of(rename_vec))) %>%
      select(c(x$variable, all_of(prereq_cols)))
  }
  out
}

#' Internal function to remove duplicated non-key variables prior to join
#'
#' This function is used with `build_from_derived` to drop columns that would
#' cause a conflict on joining datasets, prioritising keeping columns in
#' datasets earlier on in ds_list.
#'
#' e.g. if ds_list = ("AE", "ADSL") and there is a conflicting column
#' "STUDYID", the column will be dropped from ADSL (index 2) rather than AE
#' (index 1).
#'
#' @param x List of datasets with all columns added
#' @param keys List of key values to join on
#' @param ds_names Names of datasets
#' @param verbose Verbosity level
#'
#' @return datasets
#' @noRd
prepare_join <- function(x, keys, ds_names, verbose = "message") {
  out <- list(x[[1]])

  if (length(x) > 1) {
    for (i in 2:length(x)) {
      # Drop non-key cols present in each previous dataset in order
      drop_cols <- c()

      for (j in 1:(i - 1)) {
        conflicting_cols <- keep(names(x[[j]]), function(col) !(col %in% keys)) %>%
          intersect(colnames(x[[i]]))
        drop_cols <- c(drop_cols, conflicting_cols)

        if (length(conflicting_cols) > 0 && check_message(verbose)) {
          cli_inform(c("i" = "Dropping column(s) from {ds_names[[i]]} due to \
                            conflict with {ds_names[[j]]}: {conflicting_cols}."))
        }
      }

      out[[i]] <- x[[i]] %>%
        select(-any_of(drop_cols))
    }
  }
  out
}

#' Drop Unspecified Variables
#'
#' This function drops all unspecified variables.
#' @param dataset Dataset to change
#' @param metacore metacore object that only contains the specifications for the
#'   dataset of interest.
#' @param dataset_name `r lifecycle::badge("deprecated")` Optional string to specify
#' the dataset. This is only needed if the metacore object provided hasn't already
#' been subsetted.\cr
#' Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#' in a future release. Please use `metacore::select_dataset` to subset the
#' `metacore` object to obtain metadata for a single dataset.
#' @param verbose Character string controlling message verbosity. One of:
#'   \describe{
#'     \item{`"message"`}{Show both warnings and messages (default)}
#'     \item{`"warn"`}{Show warnings but suppress messages}
#'     \item{`"silent"`}{Suppress all warnings and messages}
#'   }
#'
#' @return Dataset with only specified columns
#' @export
#'
#' @examples
#' library(metacore)
#' library(haven)
#' library(dplyr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' data <- read_xpt(metatools_example("adsl.xpt")) %>%
#'   select(USUBJID, SITEID) %>%
#'   mutate(foo = "Hello")
#' drop_unspec_vars(data, spec)
drop_unspec_vars <- function(dataset, metacore, dataset_name = deprecated(),
                             verbose = c("message", "warn", "silent")) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "drop_unspec_vars(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }

  verbose <- validate_verbose(verbose)

  verify_DatasetMeta(metacore)
  var_list <- metacore$ds_vars %>%
    filter(is.na(supp_flag) | !(supp_flag)) %>%
    pull(variable)
  to_drop <- names(dataset) %>%
    discard(~ . %in% var_list)
  if (length(to_drop) > 0) {
    out <- dataset %>%
      select(-all_of(to_drop))
    if (check_message(verbose)) {
      message(paste0(
        "The following variable(s) were dropped:\n  ",
        paste0(to_drop, collapse = "\n  ")
      ))
    }
  } else {
    out <- dataset
  }
  out
}


#' Add Missing Variables
#'
#' This function adds in missing columns according to the type set in the
#' metacore object. All values in the new columns will be missing, but typed
#' correctly. If unable to recognize the type in the metacore object will return
#' a logical type.
#' @param dataset Dataset to add columns to. If all variables are present no
#'   columns will be added.
#' @param metacore metacore object that only contains the specifications for the
#'   dataset of interest.
#' @param dataset_name Optional string to specify the dataset. This is only
#'   needed if the metacore object provided hasn't already been subsetted.\cr
#'   Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#'   in a future release. Please use `metacore::select_dataset` to subset the
#'   `metacore` object to obtain metadata for a single dataset.
#'
#' @return The given dataset with any additional columns added
#' @export
#'
#' @examples
#' library(metacore)
#' library(haven)
#' library(dplyr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' data <- read_xpt(metatools_example("adsl.xpt")) %>%
#'   select(-TRTSDT, -TRT01P, -TRT01PN)
#' add_variables(data, spec)
add_variables <- function(dataset, metacore, dataset_name = deprecated()) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "add_variables(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }

  verify_DatasetMeta(metacore)
  var_list <- metacore$ds_vars %>%
    filter(is.na(supp_flag) | !(supp_flag)) %>%
    pull(variable)

  to_add <- var_list %>%
    discard(~ . %in% names(dataset))
  if (length(to_add) > 0) {
    n <- nrow(dataset)
    typing <- metacore$var_spec %>%
      filter(variable %in% to_add) %>%
      mutate(
        type_fmt = str_to_lower(type),
        out_type =
          case_when(
            str_detect(str_to_lower(format), "date") ~ "date",
            type_fmt == "integer" ~ "integer",
            type_fmt == "numeric" ~ "double",
            type_fmt == "text" ~ "character",
            type_fmt == "character" ~ "character",
            type_fmt == "boolean" ~ "logical",
            type_fmt == "logical" ~ "logical",
            TRUE ~ "unknown"
          )
      )

    new_cols <- map(typing$out_type, function(typ) {
      out <- switch(typ,
        "character" = rep(NA_character_, n),
        "integer" = rep(NA_integer_, n),
        "double" = rep(NA_real_, n),
        "date" = as.Date(rep(NA_integer_, n)),
        "logical" = rep(NA, n),
        "unknown" = rep(NA, n)
      )
    })
    names(new_cols) <- typing$variable
    new_cols <- as_tibble(new_cols)

    dataset <- bind_cols(dataset, new_cols)
  }
  dataset
}

# Suppress cli output during testing
options(cli.default_handler = function(...) {})

load(metacore::metacore_example("pilot_ADaM.rda"))
spec <- metacore %>% select_dataset("ADSL", verbose = "silent")
test_that("drop_unspec_vars", {
  data <- haven::read_xpt(metatools_example("adsl.xpt")) %>%
    mutate(AGEGR2 = "DUMMY", AGEGR2N = 99, foo = "Hello", foo2 = "world")

  man_vars <- metacore$ds_vars %>%
    filter(dataset == "ADSL") %>%
    pull(variable)
  man_dat <- data %>%
    select(all_of(man_vars))
  drop_unspec_vars(data, spec) %>%
    expect_equal(man_dat)
  expect_message(drop_unspec_vars(data, spec),
    label = "The following variable(s) were dropped:\n  foo\n  foo2"
  )
})


test_that("build_from_derived", {
  ds_list <- list(DM = haven::read_xpt(metatools_example("dm.xpt")))

  expect_error(build_from_derived(spec, ds_list, keep = FALSE))
  # Vars that should be brought through
  man_vars <- spec$derivations %>%
    filter(str_detect(derivation, "^DM\\.")) %>%
    pull(derivation) %>%
    str_remove("^DM\\.") %>%
    unique() %>%
    c("TRT01P") %>%
    sort()
  build_from_derived(spec, ds_list,
    predecessor_only = FALSE,
    keep = FALSE
  ) %>%
    names() %>%
    sort() %>%
    expect_equal(man_vars)

  # Vars pulled through with old columns kept
  man_vars <- spec$derivations %>%
    filter(str_detect(derivation, "^DM\\.")) %>%
    pull(derivation) %>%
    str_remove("^DM\\.") %>%
    unique() %>%
    c(., "TRT01P") %>%
    sort()

  expect_warning(
    build_from_derived(spec, ds_list,
      predecessor_only = FALSE,
      keep = TRUE
    ) %>%
      names() %>%
      sort() %>%
      expect_equal(man_vars),
    label = paste0(
      "Setting 'keep' = TRUE has been superseded, and will be",
      " unavailable in future releases. Please consider setting",
      " 'keep' equal to 'ALL' or 'PREREQUISITE'."
    )
  )

  # Pulling through from more than one dataset
  spec2 <- metacore %>% select_dataset("ADAE", verbose = "silent")
  adae_auto <- build_from_derived(spec2,
    ds_list = list(
      "AE" = safetyData::sdtm_ae,
      "ADSL" = safetyData::adam_adsl
    ),
    predecessor_only = FALSE,
    keep = FALSE
  )
  ae_part_vars <- spec2$derivations %>%
    filter(str_detect(derivation, "AE\\.[[:alnum:]]*$")) %>%
    pull(derivation) %>%
    str_remove("^AE\\.") %>%
    c("STUDYID", "USUBJID", .)

  ae_part <- select(safetyData::sdtm_ae, all_of(ae_part_vars))

  adsl_part_vars <- spec2$derivations %>%
    filter(str_detect(derivation, "ADSL\\.[[:alnum:]]*$")) %>%
    pull(derivation) %>%
    str_remove("^ADSL\\.")
  adsl_part <-
    select(safetyData::adam_adsl, all_of(adsl_part_vars)) |>
    rename(TRTA = TRT01A, TRTAN = TRT01AN)

  adae_man <- full_join(adsl_part, ae_part, by = c("STUDYID", "USUBJID"), multiple = "all") %>%
    select(all_of(names(adae_auto)), everything())
  expect_equal(adae_auto, adae_man)


  # Pulling through from one dataset when spec has more than one
  adae_auto_adsl_only <- build_from_derived(spec2,
    ds_list = list("ADSL" = safetyData::adam_adsl),
    predecessor_only = FALSE,
    keep = FALSE
  ) |>
    order_cols(spec2)
  adsl_man <- order_cols(adsl_part, spec2)
  expect_equal(adae_auto_adsl_only, adsl_man)

  adsl <- safetyData::adam_adsl
  ae <- safetyData::sdtm_ae
  adae_auto_unnamed <- build_from_derived(spec2,
    ds_list = list(ae, adsl),
    predecessor_only = FALSE,
    keep = FALSE
  )
  expect_equal(adae_auto, adae_man)

  expect_warning(build_from_derived(spec2,
    ds_list = list(safetyData::sdtm_ae, adsl),
    predecessor_only = FALSE,
    keep = FALSE
  ))

  # Pulling through all columns from original dataset
  adae_full <- build_from_derived(spec2,
    ds_list = list(
      "AE" = safetyData::sdtm_ae,
      "ADSL" = safetyData::adam_adsl
    ),
    predecessor_only = FALSE,
    keep = "ALL"
  )

  full_adsl_part <- safetyData::adam_adsl %>%
    mutate(TRTA = TRT01A, TRTAN = TRT01AN)

  adae_all_man <- full_join(full_adsl_part, safetyData::sdtm_ae, by = c("STUDYID", "USUBJID"), multiple = "all")

  expect_equal(adae_full, adae_all_man)

  # Pulling through columns required for future derivations
  spec3 <- metacore %>% select_dataset("ADAE", verbose = "silent")

  adae_prereq <- build_from_derived(spec3,
    ds_list = list(
      "AE" = safetyData::sdtm_ae,
      "ADSL" = safetyData::adam_adsl
    ),
    predecessor_only = FALSE,
    keep = "PREREQUISITE"
  )

  adae_auto <- build_from_derived(spec3,
    ds_list = list(
      "AE" = safetyData::sdtm_ae,
      "ADSL" = safetyData::adam_adsl
    ),
    predecessor_only = FALSE,
    keep = "PREREQUISITE"
  )


  adae_all <- build_from_derived(spec3,
    ds_list = list(
      "AE" = safetyData::sdtm_ae,
      "ADSL" = safetyData::adam_adsl
    ),
    predecessor_only = FALSE,
    keep = "ALL"
  )

  adae_prereq_man <- adae_all %>%
    select(c(names(adae_auto)))

  expect_equal(adae_prereq, adae_prereq_man)
})


test_that("add_variables add missing variables to the metacore object", {
  load(metacore::metacore_example("pilot_ADaM.rda"))
  spec <- metacore %>% select_dataset("ADSL", verbose = "silent")
  data <- haven::read_xpt(metatools_example("adsl.xpt")) %>%
    mutate(AGEGR2 = "DUMMY", AGEGR2N = 99)
  data_mis <- data %>%
    select(-TRTSDT, -TRT01P, -TRT01PN)
  # Check data when there is missing
  fx_miss <- add_variables(data_mis, spec) %>%
    select(TRTSDT, TRT01P, TRT01PN)
  man_miss <- data %>%
    mutate(
      TRTSDT = as.Date(NA_integer_),
      TRT01P = NA_character_,
      TRT01PN = NA_integer_
    ) %>%
    select(TRTSDT, TRT01P, TRT01PN)
  expect_equal(fx_miss, man_miss)
  # Check data when there isn't any missing
  expect_equal(
    add_variables(data, spec),
    data
  )
})

test_that("drop_unspec_vars verbose parameter", {
  data <- haven::read_xpt(metatools_example("adsl.xpt")) %>%
    mutate(foo = "Hello", foo2 = "world")

  man_vars <- metacore$ds_vars %>%
    filter(dataset == "ADSL") %>%
    pull(variable)
  man_dat <- data %>%
    select(all_of(man_vars))

  # Test verbose = "message" (default behavior)
  expect_message(
    drop_unspec_vars(data, spec, verbose = "message"),
    "The following variable\\(s\\) were dropped:"
  )

  # Test verbose = "warn" (suppress messages)
  expect_silent(
    drop_unspec_vars(data, spec, verbose = "warn")
  )

  # Test verbose = "silent" (suppress all output)
  expect_silent(
    drop_unspec_vars(data, spec, verbose = "silent")
  )

  # Verify all verbose levels return same result
  result_message <- drop_unspec_vars(data, spec, verbose = "message")
  result_warn <- drop_unspec_vars(data, spec, verbose = "warn")
  result_silent <- drop_unspec_vars(data, spec, verbose = "silent")

  expect_equal(result_message, man_dat)
  expect_equal(result_warn, man_dat)
  expect_equal(result_silent, man_dat)

  # Test invalid verbose value
  expect_error(
    drop_unspec_vars(data, spec, verbose = "invalid"),
    "should be one of: message, warn, silent"
  )
})

test_that("build_from_derived verbose controls prepare_join messages", {
  load(metacore::metacore_example("pilot_ADaM.rda"))
  spec2 <- metacore %>% select_dataset("ADAE", verbose = "silent")

  # Use safetyData datasets and add a conflicting non-key column
  # STUDYID is a key column, so add a different column that will conflict
  ae <- safetyData::sdtm_ae %>%
    mutate(TESTCOL = "AE_VALUE") # Add a non-key column

  adsl <- safetyData::adam_adsl %>%
    mutate(TESTCOL = "ADSL_VALUE") # Same column with different value

  ds_list <- list(AE = ae, ADSL = adsl)

  # Test that conflicting column messages are shown with verbose = "message"
  expect_message(
    build_from_derived(spec2, ds_list, predecessor_only = FALSE, verbose = "message", keep = "ALL"),
    "Dropping column\\(s\\) from"
  )

  # Test that messages are suppressed with verbose = "warn" and "silent"
  expect_silent(
    build_from_derived(spec2, ds_list, predecessor_only = FALSE, verbose = "warn", keep = "ALL")
  )

  expect_silent(
    build_from_derived(spec2, ds_list, predecessor_only = FALSE, verbose = "silent", keep = "ALL")
  )
})

#' Check Control Terminology for a Single Column
#'
#' This function checks the column in the dataset only contains the control
#' terminology as defined by the metacore specification
#'
#' @param data Data to check
#'
#' @param metacore A metacore object to get the codelist from. If the variable
#'   has different codelists for different datasets the metacore object will
#'   need to be subsetted using `select_dataset` from the metacore package.
#'
#' @param var Name of variable to check
#'
#' @param na_acceptable Logical value, set to `NULL` by default, so the
#'   acceptability of missing values is based on if the core for the variable is
#'   "Required" in the `metacore` object. If set to `TRUE` then will pass check
#'   if values are in the control terminology or are missing. If set to
#'   `FALSE`then NA will not be acceptable.
#'
#' @param verbose `character` string controlling the verbosity of the output.
#'   Possible values are `"message"` (for general information and success messages)
#'   and `"warn"` (for warnings). Partial matching is allowed.
#'   **Important**: `"silent"` is explicitly **not** a valid option for `verbose` in
#'   this function. The primary purpose of `check_ct_data` is to identify and
#'   warn the user about non-compliant or problematic control terminology. Allowing
#'   the suppression of these warnings would bypass the function's intent and could
#'   lead to unnoticed data quality issues. If `verbose = "silent"` is provided, it will
#'   be coerced to `"message"` with a warning.
#'
#' @param .internal Logical value indicating whether the function is being
#'   called internally by another package function. If `TRUE`, the function
#'   suppresses user-facing messages and instead returns a logical indicator
#'   of whether any controlled terminology violations were detected. This
#'   argument is intended for internal use only and should not be set by
#'   end users.
#'
#' @return Given data if column only contains control terms. If not, will error
#'   given the values which should not be in the column
#'
#' @export
#'
#' @examples
#' library(metacore)
#' library(haven)
#' library(magrittr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' data <- read_xpt(metatools_example("adsl.xpt"))
#' check_ct_col(data, spec, TRT01PN)
#' check_ct_col(data, spec, "TRT01PN")
check_ct_col <- function(data, metacore, var, na_acceptable = NULL, verbose = "message", .internal = FALSE) {
  verify_DatasetMeta(metacore)

  # Verbose cannot be `silent` as the point of this function is to warn the user
  tryCatch(
    {
      arg <- match.arg(verbose, "silent")
      cli_warn(c(
        "x" = "Argument {.arg verbose} cannot be {.val {arg}} for {.fn check_ct_data}",
        "i" = "Must be one of: {.val message}, {.val warn}",
        "i" = "Defaulting to {.val message}"
      ))
      verbose <- "message"
    },
    error = function(e) {
      verbose <- validate_verbose(verbose, call = rlang::env_parent())
    }
  )

  var_name <- rlang::as_name(rlang::ensym(var))

  bad_vals <- get_bad_ct(
    data = data,
    metacore = metacore,
    var = {{ var }},
    na_acceptable = na_acceptable
  )

  if (length(bad_vals) == 0) {
    if (.internal) {
      return(TRUE)
    } else if (verbose == "message") {
      return(data)
    }
    return(invisible(data))
  }

  # Format values nicely for display
  bad_vals_fmt <- paste0("'", bad_vals, "'")
  codelist <- metacore$value_spec |>
    filter(variable == var_name) |>
    pull(code_id)

  cli_warn(c(
    "x" = "Invalid controlled terminology detected",
    "i" = "Variable: {var_name} | Codelist: {codelist}",
    "i" = "Values not permitted {bad_vals_fmt}",
    ""
  ))

  invisible(FALSE)
}

#' Gets vector of control terminology which should be there
#'
#' This function checks the column in the dataset only contains the control
#' terminology as defined by the metacore specification. It will return all
#' values not found in the control terminology
#'
#' @param data Data to check
#' @param metacore A metacore object to get the codelist from. If the variable
#'   has different codelists for different datasets the metacore object will
#'   need to be subsetted using `select_dataset` from the metacore package.
#' @param var Name of variable to check
#' @param na_acceptable Logical value, set to `NULL` by default, so the
#'   acceptability of missing values is based on if the core for the variable is
#'   "Required" in the `metacore` object. If set to `TRUE` then will pass check
#'   if values are in the control terminology or are missing. If set to
#'   `FALSE` then NA will not be acceptable.
#'
#' @return vector
#' @export
#'
#' @examples
#' library(haven)
#' library(metacore)
#' library(magrittr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' data <- read_xpt(metatools_example("adsl.xpt"))
#' get_bad_ct(data, spec, "DCSREAS")
#' get_bad_ct(data, spec, "DCSREAS", na_acceptable = FALSE)
#'
get_bad_ct <- function(data, metacore, var, na_acceptable = NULL) {
  verify_DatasetMeta(metacore)
  col_name_str <- as_label(enexpr(var)) %>%
    str_remove_all("\"")
  if (!col_name_str %in% names(data)) {
    stop(paste(col_name_str, "not found in dataset. Please check and try again"), call. = FALSE)
  }
  ct <- get_control_term(metacore, {{ var }})
  if (is.vector(ct)) {
    check <- ct
  } else if ("code" %in% names(ct)) {
    check <- ct %>% pull(code)
  } else {
    stop("We currently don't have the ability to check against external libraries", call. = FALSE)
  }
  core <- metacore$ds_vars %>%
    filter(variable == col_name_str) %>%
    pull(core)
  attr(core, "label") <- NULL
  test <- ifelse(is.null(na_acceptable), !identical(core, "Required"), na_acceptable)
  if (test) {
    if (all(is.character(check))) {
      check <- c(check, NA_character_, "")
    } else {
      check <- c(check, NA)
    }
  }
  test <- pull(data, {{ var }}) %in% check
  pull(data, {{ var }})[!test] %>%
    unique()
}


#' Check Control Terminology for a Dataset
#'
#' This function checks that all columns in the dataset only contains the
#' control terminology as defined by the metacore specification.
#'
#' @param data Dataset to check
#'
#' @param metacore metacore object that contains the specifications for the
#'   dataset of interest. If any variable has different codelists for different
#'   datasets the metacore object will need to be subsetted using
#'   `select_dataset` from the metacore package.
#'
#' @param na_acceptable `logical` value or `character` vector, set to `NULL` by default.
#'   `NULL` sets the acceptability of missing values based on if the core for
#'   the variable is "Required" in the `metacore` object. If set to `TRUE` then will
#'   pass check if values are in the control terminology or are missing. If set
#'   to `FALSE` then NA will not be acceptable. If set to a `character` vector then
#'   only the specified variables may contain NA values.
#'
#' @param omit_vars `character` vector indicating which variables should be skipped
#'   when doing the controlled terminology checks. Internally, `omit_vars` is
#'   evaluated before `na_acceptable`.
#'
#' @param verbose `character` string controlling the verbosity of the output.
#'   Possible values are `"message"` (for general information and success messages)
#'   and `"warn"` (for warnings). Partial matching is allowed.
#'   **Important**: `"silent"` is explicitly **not** a valid option for `verbose` in
#'   this function. The primary purpose of `check_ct_data` is to identify and
#'   warn the user about non-compliant or problematic control terminology. Allowing
#'   the suppression of these warnings would bypass the function's intent and could
#'   lead to unnoticed data quality issues. If `verbose = "silent"` is provided, it will
#'   be coerced to `"message"` with a warning.
#'
#' @return Given data if all columns pass. It will issue a warning otherwise.
#'
#' @export
#'
#' @examples
#' library(haven)
#' library(metacore)
#' library(magrittr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL", quiet = TRUE)
#' data <- read_xpt(metatools_example("adsl.xpt"))
#'
#' check_ct_data(data, spec, omit_vars = c("AGEGR2", "AGEGR2N"))
#' \dontrun{
#' # These examples produce errors:
#' check_ct_data(data, spec, na_acceptable = FALSE)
#' check_ct_data(data, spec, na_acceptable = FALSE, omit_vars = "DISCONFL")
#' check_ct_data(data, spec, na_acceptable = c("DSRAEFL", "DCSREAS"), omit_vars = "DISCONFL")
#' }
check_ct_data <- function(data, metacore, na_acceptable = NULL, omit_vars = NULL, verbose = "message") {
  verify_DatasetMeta(metacore)

  # Verbose cannot be `silent` as the point of this function is to warn the user
  tryCatch(
    {
      arg <- match.arg(verbose, "silent")
      cli_warn(c(
        "x" = "Argument {.arg verbose} cannot be {.val {arg}} for {.fn check_ct_data}",
        "i" = "Must be one of: {.val message}, {.val warn}",
        "i" = "Defaulting to {.val message}"
      ))
      verbose <- "message"
    },
    error = function(e) {
      verbose <- validate_verbose(verbose, call = rlang::env_parent())
    }
  )

  codes_in_data <- metacore$value_spec %>%
    dplyr::filter(variable %in% names(data), !is.na(code_id)) %>%
    dplyr::pull(code_id) %>%
    unique()

  # Remove any codes that have external libraries
  codes_to_check <- metacore$codelist %>%
    dplyr::filter(type != "external_library", code_id %in% codes_in_data) %>%
    dplyr::select(code_id)

  # Convert list of codes to variables
  cols_to_check <- metacore$value_spec %>%
    dplyr::inner_join(codes_to_check, by = "code_id", relationship = "many-to-many") %>%
    dplyr::filter(variable %in% names(data)) %>%
    dplyr::pull(variable) %>%
    unique()

  # Subset cols_to_check by omit_vars
  if (is.character(omit_vars)) {
    check_vars_in_data(omit_vars, "omit_vars", data)
    cols_to_check <- setdiff(cols_to_check, omit_vars)
  }

  # Validate na_acceptable
  if (!is.null(na_acceptable) &&
    !is.logical(na_acceptable) &&
    !is.character(na_acceptable)) {
    cli::cli_abort(
      "na_acceptable must be NULL, logical, or character."
    )
  }

  # Run checks and collect flags
  results <- purrr::map_lgl(cols_to_check, function(x) {
    if (is.character(na_acceptable)) {
      na_flag <- x %in% na_acceptable
    } else if (is.logical(na_acceptable) || is.null(na_acceptable)) {
      na_flag <- na_acceptable
    }

    check_ct_col(data, metacore, !!rlang::sym(x), na_flag, "message", .internal = TRUE)
  })

  # If no warnings triggered
  if (all(results) && verbose == "message") {
    cli::cli_inform(c(
      "v" = "All controlled terminology checks passed"
    ))
  }

  # Print dataset Y/N?
  if (verbose == "message") {
    return(data)
  }

  invisible(data)
}

check_vars_in_data <- function(vars, vars_name, data) {
  if (!all(vars %in% names(data))) {
    stop(
      paste0(
        "Not all variables from ", vars_name, " are in the data: ",
        paste0(setdiff(vars, names(data)), collapse = ",")
      ),
      call. = FALSE
    )
  }
  return(NULL)
}

#' Check Variable Names
#'
#' This function checks the variables in the dataset against the variables
#' defined in the metacore specifications. If everything matches the function
#' will print a message stating everything is as expected. If there
#' are additional or missing variables an error will explain the discrepancies
#' @param data Dataset to check
#' @param metacore metacore object that only contains the specifications for the
#'   dataset of interest.
#' @param dataset_name `r lifecycle::badge("deprecated")` Optional string to
#'   specify the dataset. This is only needed if the metacore object provided
#'   hasn't already been subsetted.\cr
#'   Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#'   in a future release. Please use `metacore::select_dataset` to subset the
#'   `metacore` object to obtain metadata for a single dataset.
#' @param strict A logical value indicating whether to perform strict
#'   validation on the input dataset. If \code{TRUE} (default), errors will be raised
#'   if validation fails. If \code{FALSE}, warnings will be issued instead, allowing
#'   the function execution to continue event with invalid data.
#'
#' @return message if the dataset matches the specification and the dataset, and error otherwise
#' @export
#'
#' @examples
#' library(haven)
#' library(metacore)
#' library(magrittr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' data <- read_xpt(metatools_example("adsl.xpt"))
#' check_variables(data, spec)
#' data["DUMMY_COL"] <- NA
#' check_variables(data, spec, strict = FALSE)
check_variables <- function(data, metacore, dataset_name = deprecated(), strict = FALSE) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "check_variables(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }
  verify_DatasetMeta(metacore)

  var_list <- metacore$ds_vars %>%
    filter(is.na(supp_flag) | !(supp_flag)) %>%
    pull(variable)

  missing <- var_list %>% discard(~ . %in% names(data))
  extra <- names(data) %>% discard(~ . %in% var_list)

  messages <- character(0)
  data_list <- list()

  if (length(missing) > 0) {
    messages <- c(messages, "The following variables are missing")
    data_list <- c(data_list, list(missing))
  }

  if (length(extra) > 0) {
    messages <- c(messages, "The following variables do not belong")
    data_list <- c(data_list, list(extra))
  }

  if (length(messages) > 0) {
    print_to_console(messages, data_list, strict = {{ strict }})
  } else {
    message("No missing or extra variables")
  }

  data
}

#' Print Messages to Console
#'
#' This function prints formatted messages to the console, either as errors (stopping
#' execution) or as warnings. It is designed as a helper function to provide informative
#' messages during validation checks.
#'
#' @param messages A character vector of messages to be printed. Each element corresponds
#'   to a separate message.
#' @param data_list A list of character vectors. Each element in the list corresponds
#'   to a message in `messages` and provides associated data (e.g., column names).
#'   If an element in `messages` has no corresponding data, include a `NULL`.
#' @param strict A logical value indicating whether to print messages as
#'   errors (\code{TRUE}, default) or warnings (\code{FALSE}).
#'
#' @details The function constructs a formatted message string including the calling
#' function's name, the individual messages provided in `messages`, and associated data
#' from `data_list`. The function uses \code{switch} to call either `stop()` or `warning()`
#' based on `strict` and prints the full message string to the console.
#'
#' @return None. The function's primary purpose is its side effect of printing a message.
#' It does not return a meaningful value.
#'
#' @noRd
#'
print_to_console <- function(messages, data_list, strict = TRUE) {
  calling_function <- paste(deparse(sys.call(-1)), collapse = " ")
  output_string <- paste0("In: [", calling_function, "]")

  for (i in seq_along(messages)) {
    message <- paste0(messages[i], ": ",
      paste(data_list[[i]], collapse = ", "),
      sep = "\n"
    )

    output_string <- paste(output_string, message, sep = "\n\n")
  }

  options(deparse.max.lines = 2000L)
  switch(as.character(strict),
    "TRUE"  = cli::cli_abort(c(output_string), call = NULL),
    "FALSE" = cli::cli_warn(c(output_string), call = NULL)
  )
}

#' Check Uniqueness of Records by Key
#'
#' This function checks the uniqueness of records in the dataset by key using
#' `get_keys` from the metacore package. If the key uniquely identifies each
#' record the function will print a message stating everything is as expected.
#' If records are not uniquely identified an error will explain the duplicates.
#' @param data Dataset to check
#' @param metacore metacore object that only contains the specifications for the
#'   dataset of interest.
#' @param dataset_name `r lifecycle::badge("deprecated")` Optional string to
#'   specify the dataset that is being built. This is only needed if the metacore
#'   object provided hasn't already been subsetted.\cr
#'   Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#'   in a future release. Please use `metacore::select_dataset` to subset the
#'   `metacore` object to obtain metadata for a single dataset.
#'
#' @return message if the key uniquely identifies each dataset record, and error otherwise
#' @export
#'
#' @examples
#' library(haven)
#' library(metacore)
#' library(magrittr)
#' load(metacore_example("pilot_ADaM.rda"))
#' spec <- metacore %>% select_dataset("ADSL")
#' data <- read_xpt(metatools_example("adsl.xpt"))
#' check_unique_keys(data, spec)
check_unique_keys <- function(data, metacore, dataset_name = deprecated()) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "check_unique_keys(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }
  verify_DatasetMeta(metacore)
  keys <- get_keys(metacore, expr(!!metacore$ds_spec$dataset))
  var_list <- keys %>%
    pull(variable)
  missing <- var_list %>%
    discard(~ . %in% names(data))
  if (length(missing) > 0) {
    stop(paste0(
      "The following variable keys are missing in the dataset:\n",
      paste0(missing, collapse = "\n")
    ))
  }
  grouped <- data %>%
    group_by(pick(!!keys$variable)) %>%
    add_count() %>%
    filter(.data$n != 1)
  if (nrow(grouped) == 0) {
    message("Keys uniquely identify records")
  } else {
    stop(paste0(
      "Keys do not uniquely identify records\n",
      "variable keys:\n",
      paste0(var_list, collapse = "\n")
    ))
  }
  data
}

# Suppress cli output during testing
options(cli.default_handler = function(...) {})

# Load data to use across tests
load(metacore::metacore_example("pilot_ADaM.rda"))
spec <- metacore %>% select_dataset("ADSL", verbose = "silent")
data <- haven::read_xpt(metatools_example("adsl.xpt"))

mod_ds_vars <- spec$ds_vars %>%
  mutate(core = if_else(variable %in% c("TRT01PN", "COMP8FL"), "Required", core))
spec_mod <- metacore::metacore(spec$ds_spec, mod_ds_vars, spec$var_spec, spec$value_spec, spec$derivations, spec$codelist, verbose = "silent")
spec_mod <- select_dataset(spec_mod, "ADSL", verbose = "silent")

test_that("get_bad_ct works correctly", {
  # test na_acceptable
  expect_equal(get_bad_ct(data, spec, "DCSREAS"), character(0))
  expect_equal(get_bad_ct(data, spec, "DCSREAS", TRUE), character(0))
  expect_equal(get_bad_ct(data, spec, "DCSREAS", FALSE), "")

  expect_equal(get_bad_ct(data, spec_mod, "COMP8FL"), "")
  expect_equal(get_bad_ct(data, spec_mod, "COMP8FL", TRUE), character(0))
  expect_equal(get_bad_ct(data, spec_mod, "COMP8FL", FALSE), "")

  data_na <- data %>%
    mutate(COMP8FL = if_else(dplyr::row_number() == 1, NA_character_, COMP8FL))
  expect_equal(get_bad_ct(data_na, spec_mod, "COMP8FL"), c(NA_character_, ""))
})

test_that("check_ct_col works correctly", {
  # Check it works with a character col
  expect_equal(check_ct_col(data, spec, ARM), data)
  # Check it works with a numeric col
  expect_equal(check_ct_col(data, spec, TRT01PN), data)
  # Check it works when passes a string
  expect_equal(check_ct_col(data, spec, "TRT01PN"), data)

  # Check verbose parameters work (warn) and partial match (w)
  expect_silent(check_ct_col(data, spec, ARM, verbose = "warn"))
  expect_silent(check_ct_col(data, spec, ARM, verbose = "w"))

  # Check verbose warning issued when `verbose = "silent"` or partial match "s"
  expect_warning(check_ct_col(data, spec, ARM, verbose = "silent"))
  expect_warning(check_ct_col(data, spec, ARM, verbose = "s"))

  # Test permitted Values
  spec2 <- metacore::spec_to_metacore(metacore::metacore_example("p21_mock.xlsx"), verbose = "silent")
  dm <- select_dataset(spec2, "DM", verbose = "silent")
  expect_equal(check_ct_col(data, dm, ARM), data)

  # Test external dictionaries
  data2 <- tibble::tibble(AELLT = "Hello")
  ae <- select_dataset(spec2, "AE", verbose = "silent")
  expect_error(
    check_ct_col(data2, ae, AELLT),
    "We currently don't have the ability to check against external libraries"
  )

  # Test a column that isn't in the dataset
  expect_error(
    check_ct_col(data, ae, AELLT),
    "AELLT not found in dataset. Please check and try again"
  )

  # Test NA acceptable
  expect_error(check_ct_col(data, dm, COMP8FL, FALSE))

  expect_equal(check_ct_col(data, dm, ARM, TRUE), data)
  data_w_miss <- data %>%
    mutate(TRT01PN = if_else(dplyr::row_number() == 3, NA_real_, TRT01PN))
  expect_warning(check_ct_col(data_w_miss, spec, TRT01PN, FALSE))
  expect_equal(get_bad_ct(data_w_miss, spec, TRT01PN, FALSE), NA_real_)
  expect_equal(check_ct_col(data_w_miss, spec, TRT01PN, TRUE), data_w_miss)
  ### Test with  a required column ###
  # Required without missing
  expect_equal(check_ct_col(data, spec_mod, TRT01PN), data)
  # Required with missing
  expect_warning(check_ct_col(data, spec_mod, COMP8FL))
  expect_equal(get_bad_ct(data, spec_mod, COMP8FL), "")
  expect_equal(check_ct_col(data, spec_mod, COMP8FL, TRUE), data)
})

test_that("check_ct_data works correctly", {
  # Checking error for multiple words in multiple columns
  data_multi_word <- data %>%
    mutate(
      TRT01P = case_when(
        dplyr::row_number() == 2 ~ "Hello",
        dplyr::row_number() == 3 ~ "World",
        TRUE ~ TRT01P
      ),
      TRT01A = TRT01P
    )
  expect_warning(check_ct_data(data_multi_word |> select(TRT01P), spec))

  expect_warning(check_ct_data(data |> select(COMP8FL), spec, FALSE))

  # Check data returned
  expect_equal(check_ct_data(data, spec, omit_vars = c("AGEGR2", "AGEGR2N")), data)
  expect_equal(check_ct_data(data, spec, TRUE, omit_vars = c("AGEGR2", "AGEGR2N")), data)
  expect_warning(check_ct_data(data |> select(AGEGR2), spec_mod))
  expect_equal(check_ct_data(data, spec_mod, TRUE, omit_vars = c("AGEGR2", "AGEGR2N")), data)

  # Check verbose parameter
  expect_message(
    {
      ret <- check_ct_data(data, spec, omit_vars = c("AGEGR2", "AGEGR2N"), verbose = "message")
    },
    regexp = "All controlled terminology checks passed"
  )

  expect_silent(check_ct_data(data, spec, omit_vars = c("AGEGR2", "AGEGR2N"), verbose = "warn"))
  expect_silent(check_ct_data(data, spec, omit_vars = c("AGEGR2", "AGEGR2N"), verbose = "w"))

  ret <- check_ct_data(data, spec, omit_vars = c("AGEGR2", "AGEGR2N"), verbose = "silent") |>
    expect_message("All controlled terminology checks passed") |>
    expect_warning("Argument")

  ret <- check_ct_data(data, spec, na_acceptable = c("DCSREAS", "COMP8FL", "BMIBLGR1"), verbose = "silent") |>
    expect_warning("Argument") |>
    expect_warning("Invalid controlled terminology detected")

  # Check character vector input for na_acceptable:
  expect_warning(check_ct_data(data, spec, na_acceptable = c("DCSREAS", "COMP8FL", "BMIBLGR1")))
  expect_error(check_ct_data(data, spec, 1))

  # Check omit_vars:
  expect_error(check_ct_data(data, spec, omit_vars = c("A", "B")))
  expect_warning(check_ct_data(data, spec, FALSE, omit_vars = c("DCSREAS", "COMP8FL", "BMIBLGR1")))
  expect_equal(
    check_ct_data(
      data,
      spec_mod,
      na_acceptable = NULL,
      omit_vars = c("AGEGR2", "AGEGR2N", "COMP8FL")
    ),
    data
  )
})

test_that("variable_check works correctly", {
  expect_equal(check_variables(data, spec), data)
  data_miss <- data %>% select(-1)
  expect_error(check_variables(data_miss, spec, strict = TRUE))
  expect_warning(check_variables(data_miss, spec, strict = FALSE))
  data_extra <- data %>% mutate(foo = "hello")
  expect_error(check_variables(data_extra, spec, strict = TRUE))
  expect_warning(check_variables(data_extra, spec, strict = FALSE))
  data_mis_ex <- data_extra %>% select(-1)
  expect_error(check_variables(data_mis_ex, spec, strict = TRUE))
  expect_warning(check_variables(data_mis_ex, spec, strict = FALSE))
})

test_that("check_unique_keys works as expected", {
  # check requirement for subsetted metacore object or a dataset name
  expect_error(check_unique_keys(data, metacore))
  # check missing variable keys error
  adae <- select_dataset(metacore, "ADAE", verbose = "silent")
  expect_error(check_unique_keys(data, adae))
  # check works correctly when records are unique
  adsl <- select_dataset(metacore, "ADSL", verbose = "silent")
  expect_message(check_unique_keys(data, adsl))
  # check works correctly when records are not unique
  test <- build_from_derived(adae,
    ds_list = list(
      "AE" = safetyData::sdtm_ae,
      "ADSL" = safetyData::adam_adsl
    ),
    predecessor_only = FALSE,
    keep = FALSE
  )
  expect_error(check_unique_keys(test, adae))
})

#' Build the observations for a single QNAM
#'
#' @param dataset Input dataset
#' @param qnam  QNAM value
#' @param qlabel QLABEL value
#' @param idvar IDVAR variable name (provided as a string)
#' @param qeval QEVAL value to be populated for this QNAM
#' @param qorig QORIG value to be populated for this QNAM
#' @param verbose Character string controlling message verbosity. One of:
#'   \describe{
#'     \item{`"message"`}{Show both warnings and messages (default)}
#'     \item{`"warn"`}{Show warnings but suppress messages}
#'     \item{`"silent"`}{Suppress all warnings and messages}
#'   }
#'
#' @return Observations structured in SUPP format
#' @export
#'
build_qnam <- function(dataset, qnam, qlabel, idvar, qeval, qorig,
                       verbose = c("message", "warn", "silent")) {
  verbose <- validate_verbose(verbose)

  # Need QNAM as a variable
  qval <- as.symbol(qnam)

  # DM won't have an IDVAR so handle that
  if (is.na(idvar) || idvar == "") {
    dataset <- dataset %>%
      mutate(IDVARVAL = idvar)
    idvarval <- sym("IDVARVAL")
  } else {
    idvarval <- as.symbol(idvar)
  }

  dup_sup <- dataset %>%
    select(STUDYID, RDOMAIN = DOMAIN, USUBJID, !!idvarval, !!qval) %>%
    rename(IDVARVAL = !!idvarval, QVAL = !!qval) %>%
    filter(!is.na(QVAL)) %>%
    mutate(
      IDVAR = idvar,
      QNAM = qnam,
      QLABEL = qlabel,
      QORIG = qorig,
      QEVAL = qeval
    )

  out <- dup_sup %>%
    distinct(STUDYID, RDOMAIN,
      USUBJID, IDVARVAL, QNAM,
      .keep_all = TRUE
    ) %>%
    select(
      STUDYID, RDOMAIN, USUBJID, IDVAR,
      IDVARVAL, QNAM, QLABEL, QVAL,
      QORIG, QEVAL
    )

  test_out <- dup_sup %>%
    distinct()
  if (nrow(out) != nrow(test_out)) {
    stop("The combination of STUDYID, RDOMAIN, USUBJID, IDVARVAL, QNAM is ambiguous. Consider modifying the IDVAR",
      call. = FALSE
    )
  }

  blank_test <- out %>%
    pull(QVAL)
  if (any(blank_test == "")) {
    if (check_message(verbose)) {
      message(paste0("Empty QVAL rows removed for QNAM = ", unique(out$QNAM)))
    }
    out <- out %>%
      filter(QVAL != "")
  }
  out
}


#' Make Supplemental Qualifier
#'
#' @param dataset dataset the supp will be pulled from
#' @param metacore A subsetted metacore object to get the supp information from.
#'   If not already subsetted then a `dataset_name` will need to be provided
#' @param dataset_name `r lifecycle::badge("deprecated")` Optional string to
#'   specify the dataset that is being built. This is only needed if the metacore
#'   object provided hasn't already been subsetted.\cr
#'   Note: Deprecated in version 0.2.0. The `dataset_name` argument will be removed
#'   in a future release. Please use `metacore::select_dataset` to subset the
#'   `metacore` object to obtain metadata for a single dataset.
#'
#' @return a CDISC formatted SUPP dataset
#' @export
#'
#' @examples
#'
#' library(metacore)
#' library(safetyData)
#' library(tibble)
#' load(metacore_example("pilot_SDTM.rda"))
#' spec <- metacore %>% select_dataset("AE")
#' ae <- combine_supp(sdtm_ae, sdtm_suppae)
#' make_supp_qual(ae, spec) %>% as_tibble()
make_supp_qual <- function(dataset, metacore, dataset_name = deprecated()) {
  if (is_present(dataset_name)) {
    lifecycle::deprecate_warn(
      when = "0.2.0",
      what = "make_supp_qual(dataset_name)",
      details = cli_inform(c("i" = col_red("The {.arg dataset_name} argument will be removed in a future release.
      Please use {.fn metacore::select_dataset} to subset the {.obj metacore} object to obtain
      metadata for a single dataset.")))
    )
    metacore <- make_lone_dataset(metacore, dataset_name)
  }
  verify_DatasetMeta(metacore)

  supp_vars <- metacore$ds_vars %>%
    filter(supp_flag)
  if (nrow(supp_vars) == 0) {
    stop("No supplemental variables specified in metacore object. Please check your specifications",
      call. = FALSE
    )
  }

  supp_meta <- supp_vars %>%
    select(dataset, variable) %>%
    left_join(metacore$var_spec, by = "variable") %>%
    left_join(metacore$value_spec, by = c("dataset", "variable")) %>%
    left_join(metacore$supp, by = c("dataset", "variable")) %>%
    select(
      qnam = variable, qlabel = label,
      qorig = origin, qeval = qeval,
      idvar = idvar
    ) %>%
    distinct() # Protection against bad specs
  # TODO Addin in checks/coercion for when combining cols of different types
  pmap_dfr(supp_meta, build_qnam, dataset = dataset) %>%
    arrange(USUBJID, QNAM, IDVARVAL)
}

#' Combine the Domain and Supplemental Qualifier
#'
#' @param dataset Domain dataset
#' @param supp Supplemental Qualifier dataset
#'
#' @return a dataset with the supp variables added to it
#' @export
#
#' @examples
#' library(safetyData)
#' library(tibble)
#' combine_supp(sdtm_ae, sdtm_suppae) %>% as_tibble()
combine_supp <- function(dataset, supp) {
  if (!is.data.frame(dataset) | !is.data.frame(supp)) {
    stop("You must supply a domain and supplemental dataset", call. = FALSE)
  }
  if (nrow(supp) == 0) {
    warning("Zero rows in supp, returning original dataset unchanged")
    return(dataset)
  }

  # Verify required dataset cols are present
  required_vars <- c("STUDYID", "DOMAIN", "USUBJID")
  missing_vars <- setdiff(required_vars, names(dataset))

  if (length(missing_vars) > 0) {
    cli::cli_abort(c(
      "x" = "Core SDTM variables are missing from the dataset:",
      "i" = "{.val {missing_vars}}"
    ))
  }


  # Verify required supp cols are present
  supp_cols <- c(
    "STUDYID", "RDOMAIN", "USUBJID", "IDVAR", "IDVARVAL",
    "QNAM", "QLABEL", "QVAL", "QORIG"
  )
  maybe <- c("QEVAL")

  ext_supp_col <- setdiff(names(supp), c(supp_cols, maybe))
  mis_supp_col <- setdiff(supp_cols, names(supp))

  if (length(ext_supp_col) > 0 || length(mis_supp_col) > 0) {
    cli::cli_abort(c(
      "x" = "Supplemental Qualifier dataset does not comply with CDISC SDTM structure.",
      if (length(ext_supp_col) > 0) {
        c(
          "!" = "Unexpected columns detected (must be removed):",
          "i" = "{.val {ext_supp_col}}"
        )
      },
      if (length(mis_supp_col) > 0) {
        c(
          "!" = "Required columns are missing:",
          "i" = "{.val {mis_supp_col}}"
        )
      }
    ))
  }

  # Verify qnam values from supp do not conflict with main dataset
  all_qnam <- unique(supp$QNAM)
  existing_qnam <- intersect(all_qnam, names(dataset))

  if (length(existing_qnam) > 0) {
    cli::cli_abort(c(
      "x" = "Column name conflict detected when combining SUPP data.",
      "!" = "The following QNAM values would create variables that already exist in the dataset:",
      "x" = "{.val {existing_qnam}}",
      "i" = "Renaming or removing these SUPP qualifiers is required before merging."
    ))
  }

  # In order to prevent issues when there are multiple IDVARS we need to merge
  # each IDVAR into the domain separately (otherwise there is problems when the
  # two IDVARS don't overlap)
  supp_wides_prep <-
    supp %>%
    select(-any_of(c("QLABEL", "QORIG", "QEVAL"))) %>% # Removing columns not for the main dataset
    rename(DOMAIN = RDOMAIN) %>%
    group_by(IDVAR, QNAM) %>% # For when there are multiple IDs
    group_split()

  supp_wides <- purrr::pmap(.l = list(supp = supp_wides_prep), .f = combine_supp_make_wide)

  # Verify that each idvar in supp domain is present in the main dataset
  idvars <- sapply(supp_wides, function(x) x$IDVAR[1])
  invalid_idvars <- unique(
    idvars[
      !idvars %in% names(dataset) &
        !is.na(idvars) &
        idvars != "NA"
    ]
  )

  if (length(invalid_idvars) > 0) {
    supp_wides <- supp_wides[
      vapply(supp_wides, function(x) !(x$IDVAR[1] %in% invalid_idvars), logical(1))
    ]

    cli::cli_warn(c(
      "!" = "The following {.field IDVAR} values from the SUPP dataset will not be joined:",
      "x" = "{.val {invalid_idvars}}",
      "i" = "They do not exist as variables in the main dataset."
    ))
  }

  # If all idvars are invalid then return the main dataset
  if (length(supp_wides) == 0) {
    return(dataset)
  }

  ret <- reduce(.x = append(list(dataset), supp_wides), .f = combine_supp_join)
  ret$IDVARVAL <- NULL

  labels_to_add <- unique(supp[, c("QNAM", "QLABEL")])
  for (current_idx in seq_len(nrow(labels_to_add))) {
    current_col <- labels_to_add$QNAM[current_idx]
    current_label <- labels_to_add$QLABEL[current_idx]
    attr(ret[[current_col]], "label") <- current_label
  }

  ret
}

# Create a wide version of `supp` for merging into the source dataset.
combine_supp_make_wide <- function(supp) {
  stopifnot(length(unique(supp$IDVAR)) == 1)
  stopifnot(length(unique(supp$QNAM)) == 1)
  # Get the IDVAR value to allow for renaming of IDVARVAL
  id_var <- unique(supp$IDVAR)

  wide_x <-
    supp %>%
    pivot_wider(
      names_from = QNAM,
      values_from = QVAL
    )
  wide_x$QNAM <- unique(supp$QNAM)
  if (!is.na(id_var) && id_var != "") {
    wide_x <-
      wide_x %>%
      mutate(IDVARVAL = str_trim(as.character(IDVARVAL)))
  } else {
    wide_x$IDVARVAL <- NULL
  }
  wide_x
}

combine_supp_join <- function(dataset, supp) {
  current_idvar <- unique(supp$IDVAR)
  current_qnam <- unique(supp$QNAM)
  stopifnot(length(current_idvar) == 1)
  stopifnot(length(current_qnam) == 1)

  by <- intersect(names(supp), c("STUDYID", "DOMAIN", "USUBJID", "IDVARVAL"))
  supp_prep <- supp %>% select(-QNAM, -IDVAR)
  new_column <- setdiff(names(supp_prep), by)
  stopifnot(length(new_column) == 1)

  # Prepare IDVARVAL
  ret <- dataset
  if ("IDVARVAL" %in% by) {
    # Match the IDVARVAL column in supp
    ret$IDVARVAL <- str_trim(as.character(ret[[current_idvar]]))
  } else {
    # A dummy column that can be removed later
    ret$IDVARVAL <- FALSE
  }

  # Put the new data in
  if (new_column %in% names(dataset)) {
    # Patch the data
    mask_na_ret_before <- is.na(ret[[new_column]])
    ret_orig <- ret
    ret <- dplyr::rows_patch(x = ret, y = supp_prep, by = by)
    mask_na_ret_after <- is.na(ret[[new_column]])

    expected_na_difference <- sum(!is.na(supp_prep[[new_column]]))
    actual_na_difference <- sum(!mask_na_ret_after) - sum(!mask_na_ret_before)
    if (expected_na_difference != actual_na_difference) {
      cli::cli_abort(c(
        "X" = "SUPP domain merge failed due to inconsistent key mapping.",
        "i" = "While processing {.field QNAM} = {.val {current_qnam}} with {.field IDVAR} = {.val {current_idvar}}.",
        "i" = "This usually indicates that multiple records in the SUPP domain map to the same parent record.",
        "i" = "Each combination of STUDYID, USUBJID, IDVAR, and IDVARVAL should uniquely identify a row.",
        "i" = "Check for duplicate or conflicting mappings in SUPP-- for this QNAM."
      ), call = rlang::caller_env(n = 3)) # caller_env depth of 3 is "combine_supp"
    }
  } else {
    # Verify that nothing will be missed
    missing <- dplyr::anti_join(supp_prep, ret, by = by)

    # Add a message for when there are rows in the dataset that didn't get merged
    if (nrow(missing) > 0) {
      missing_display <- missing %>%
        dplyr::transmute(
          USUBJID,
          !!current_idvar := IDVARVAL
        )

      cli::cli_warn(c(
        "x" = "Some SUPP records were not merged into the main dataset.",
        "!" = "Unmatched rows:",
        sprintf(
          "{.strong USUBJID}: {.val %s}  |  {.strong %s}: {.val %s}",
          missing_display$USUBJID,
          current_idvar,
          missing_display[[current_idvar]]
        )
      ))
    }

    # join the data
    ret <- left_join(ret, supp_prep, by = by)
  }
  ret
}

#' Handles the combining of datasets and supps for a single IDVAR
#'
#' @param dataset Domain dataset
#' @param supp Supplemental Qualifier dataset with a single IDVAR
#'
#' @return list of datasets
#' @noRd
combine_supp_by_idvar <- function(dataset, supp) {
  # Get the IDVAR value to allow for renaming of IDVARVAL
  id_var <- unique(supp$IDVAR)

  wide_x <- supp %>%
    pivot_wider(
      names_from = QNAM,
      values_from = QVAL
    ) %>%
    select(-IDVAR)

  if (!is.na(id_var) && id_var != "") {
    id_var_sym <- sym(id_var)

    by <- c("STUDYID", "DOMAIN", "USUBJID", "IDVARVAL")
    wide_x <- wide_x %>%
      mutate(IDVARVAL = as.character(IDVARVAL) %>%
        str_trim())
    #  Make a dummy IDVARVAL variable to merge on, won't effect the dataset
    dataset_chr <- dataset %>%
      mutate(IDVARVAL = as.character(!!id_var_sym) %>%
        str_trim())

    out <- left_join(dataset_chr, wide_x,
      by = by
    ) %>%
      select(-IDVARVAL)
    missing <- anti_join(wide_x, dataset_chr, by = by)

    # Add message for when there are rows in the supp that didn't get merged
    if (nrow(missing) > 0) {
      missing_txt <- capture.output(missing %>%
        select(USUBJID, !!sym(id_var)) %>%
        print()) %>%
        paste0(collapse = "\n")
      stop(
        paste0(
          "Not all rows of the Supp were merged. The following rows are missing:\n",
          missing_txt
        ),
        call. = FALSE
      )
    }
  } else {
    wide_x <- wide_x %>%
      select(-IDVARVAL)
    out <- left_join(dataset, wide_x,
      by = c("STUDYID", "DOMAIN", "USUBJID")
    )
  }
  out
}

# Suppress cli output during testing
options(cli.default_handler = function(...) {})

test_that("build_qnam", {
  full_ae <- safetyData::sdtm_suppae %>%
    select(-QORIG, -QEVAL, -QLABEL) %>%
    pivot_wider(names_from = QNAM, values_from = QVAL) %>%
    rename(AESEQ = IDVARVAL) %>%
    select(-IDVAR) %>%
    left_join(safetyData::sdtm_ae, ., by = c("STUDYID", "USUBJID", "AESEQ"))

  supp_fx <- build_qnam(
    full_ae, "AETRTEM", "TREATMENT EMERGENT FLAG",
    "AESEQ", "CLINICAL STUDY SPONSOR", "DERIVED"
  ) %>%
    select(STUDYID, RDOMAIN, USUBJID, IDVAR, IDVARVAL, QNAM, QLABEL, QVAL, QORIG, QEVAL) %>%
    arrange(USUBJID, IDVARVAL)
  ex_supp <- arrange(safetyData::sdtm_suppae, USUBJID, IDVARVAL)
  # Test standard example
  expect_equal(supp_fx, ex_supp)
  # Test without IDVAR making ambiguous output
  expect_error(
    build_qnam(
      full_ae, "AETRTEM", "TREATMENT EMERGENT FLAG",
      "", "CLINICAL STUDY SPONSOR", "DERIVED"
    ),
    "The combination of STUDYID, RDOMAIN, USUBJID, IDVARVAL, QNAM is ambiguous. Consider modifying the IDVAR"
  )
  # Test without IDVAR
  supp_sans_id <- full_ae %>%
    group_by(USUBJID) %>%
    arrange(AESEQ) %>%
    dplyr::slice(1) %>%
    build_qnam(
      "AETRTEM", "TREATMENT EMERGENT FLAG",
      "", "CLINICAL STUDY SPONSOR", "DERIVED"
    ) %>%
    select(STUDYID, RDOMAIN, USUBJID, IDVAR, IDVARVAL, QNAM, QLABEL, QVAL, QORIG, QEVAL) %>%
    arrange(USUBJID, IDVARVAL)
  ex_supp_sans_id <- arrange(safetyData::sdtm_suppae, USUBJID, IDVARVAL) %>%
    group_by(USUBJID) %>%
    dplyr::slice(1) %>%
    mutate(IDVAR = "", IDVARVAL = "")
  expect_equal(supp_sans_id, ex_supp_sans_id)
})

test_that("make_supp_qual", {
  load(metacore::metacore_example("pilot_SDTM.rda"))

  spec <- metacore %>%
    select_dataset("AE", verbose = "silent")

  # Add the mock supp variables
  ae <- combine_supp(safetyData::sdtm_ae, safetyData::sdtm_suppae)

  metacore_supp <- make_supp_qual(ae, spec) %>%
    arrange(USUBJID, QNAM, IDVARVAL) %>%
    as_tibble()

  man_supp <- ae %>%
    select(STUDYID, USUBJID, RDOMAIN = DOMAIN, IDVARVAL = AESEQ, AETRTEM) %>%
    tidyr::pivot_longer(AETRTEM, names_to = "QNAM", values_to = "QVAL") %>%
    filter(!is.na(QVAL)) %>%
    mutate(
      IDVAR = "AESEQ",
      QORIG = "derived",
      QEVAL = "CLINICAL STUDY SPONSOR",
      QLABEL = "TREATMENT EMERGENT FLAG"
    ) %>%
    arrange(USUBJID, QNAM, IDVARVAL) %>%
    select(
      STUDYID, RDOMAIN, USUBJID, IDVAR,
      IDVARVAL, QNAM, QLABEL, QVAL, QORIG, QEVAL
    ) %>%
    distinct()

  # Testing normal circumstances
  expect_equal(metacore_supp, man_supp)

  # Add the supp without a idvar
  dm_spec <- select_dataset(metacore, "DM", verbose = "silent")
  dm <- combine_supp(safetyData::sdtm_dm, safetyData::sdtm_suppdm) %>%
    as_tibble()
  dm_supp <- make_supp_qual(dm, dm_spec)
  man_dm_supp <- safetyData::sdtm_suppdm %>%
    as_tibble() %>%
    mutate(
      IDVAR = as.character(IDVAR),
      IDVARVAL = as.character(IDVARVAL),
      QORIG = tolower(QORIG)
    ) %>%
    select(STUDYID, RDOMAIN, USUBJID, IDVAR, IDVARVAL, QNAM, QLABEL, QVAL, QORIG, QEVAL)
  expect_equal(dm_supp, man_dm_supp)

  # Testing with blank rows
  supp_with_miss <- dm %>%
    dplyr::bind_rows(tibble::tibble(
      STUDYID = "CDISCPILOT01",
      DOMAIN = "DM",
      USUBJID = "01-701-9999",
      SUBJID = 9999,
      ITT = ""
    ))
  expect_message(
    make_supp_qual(supp_with_miss, dm_spec),
    "Empty QVAL rows removed for QNAM = ITT"
  )

  suppressMessages(make_supp_qual(supp_with_miss, dm_spec)) %>%
    expect_equal(man_dm_supp)


  # Testing with too many datasets
  expect_error(make_supp_qual(ae, metacore))
  # Testing without supp columns specified
  metacore_old <- metacore::spec_to_metacore(metacore::metacore_example("SDTM_spec_CDISC_pilot.xlsx"), verbose = "silent")
  ae_spec <- select_dataset(metacore_old, "AE", verbose = "silent")
  expect_error(
    make_supp_qual(ae, ae_spec),
    "No supplemental variables specified in metacore object. Please check your specifications"
  )
})


test_that("combine_supp", {
  ### 1 IDVAR and 1 QNAM
  combo_ae <- combine_supp(safetyData::sdtm_ae, safetyData::sdtm_suppae) %>%
    select(USUBJID, AESEQ, AETRTEM) %>%
    distinct() %>%
    arrange(USUBJID, AESEQ)
  supp_check <- safetyData::sdtm_suppae %>%
    select(USUBJID, AESEQ = IDVARVAL, AETRTEM = QVAL) %>%
    arrange(USUBJID, AESEQ)
  attr(supp_check$AETRTEM, "label") <- "TREATMENT EMERGENT FLAG"
  expect_equal(combo_ae, supp_check)

  ### No IDVAR and multiple QNAM
  out_test <- safetyData::sdtm_suppdm %>%
    filter(USUBJID %in% c("01-701-1015")) %>%
    select(USUBJID, QNAM, QVAL) %>%
    pivot_wider(names_from = QNAM, values_from = QVAL) %>%
    as.data.frame()
  attr(out_test$COMPLT16, "label") <- "Completers of Week 16 Population Flag"
  attr(out_test$COMPLT24, "label") <- "Completers of Week 24 Population Flag"
  attr(out_test$COMPLT8, "label") <- "Completers of Week 8 Population Flag"
  attr(out_test$EFFICACY, "label") <- "Efficacy Population Flag"
  attr(out_test$ITT, "label") <- "Intent to Treat Population Flag"
  attr(out_test$SAFETY, "label") <- "Safety Population Flag"

  full_dm <- combine_supp(safetyData::sdtm_dm, safetyData::sdtm_suppdm) %>%
    select(USUBJID, COMPLT16:SAFETY)
  expect_equal(filter(full_dm, USUBJID == "01-701-1015"), out_test)
  # Test SUBJID that wasn't in the SUPP that all supp values are NA
  full_dm %>%
    filter(USUBJID == "01-701-1057") %>%
    select(-USUBJID) %>%
    tidyr::pivot_longer(everything()) %>%
    dplyr::summarise(test = all(is.na(value))) %>%
    expect_equal(tibble::tibble(test = TRUE))

  ### Where there are only value for a small number of subjects
  mostly_miss <- combine_supp(safetyData::sdtm_ds, safetyData::sdtm_suppds)
  original <- safetyData::sdtm_suppds %>%
    arrange(USUBJID) %>%
    pull(QVAL)
  attr(original, "label") <- "PROTOCOL ENTRY CRITERIA NOT MET"
  expect_equal(
    mostly_miss %>%
      filter(!is.na(ENTCRIT)) %>%
      arrange(USUBJID) %>%
      pull(ENTCRIT),
    original
  )

  ### Multiple IDVARS and multiple QNAMS
  # Add some mock supp variables
  ae <- safetyData::sdtm_ae %>%
    mutate(
      SUPPVAR1 = letters[1:nrow(safetyData::sdtm_ae)],
      SUPPVAR2 = rep(letters, 36)[1:nrow(safetyData::sdtm_ae)],
      SUPPVAR3 = USUBJID,
      IDVAR = as.numeric(str_extract(USUBJID, "\\d{3}$"))
    )
  attr(ae$SUPPVAR1, "label") <- "Supp Test 1"
  attr(ae$SUPPVAR2, "label") <- "Supp Test 2"
  attr(ae$SUPPVAR3, "label") <- "Supp Test 3"
  ### Mock up a metadata necessary to make the SUPP
  supp_meta <- tibble::tribble(
    ~qnam, ~qlabel, ~idvar, ~qeval, ~qorig,
    "SUPPVAR1", "Supp Test 1", "AESEQ", "Investigator", "CRF",
    "SUPPVAR2", "Supp Test 2", "AESEQ", "Investigator", "CRF",
    "SUPPVAR3", "Supp Test 3", "IDVAR", "Investigator", "CRF",
  )

  ### Wrap and map
  suppae <- pmap_dfr(supp_meta, build_qnam, dataset = ae) %>%
    arrange(USUBJID, QNAM, IDVARVAL)

  dataset <- ae %>%
    select(-starts_with("SUPP"))
  supp <- suppae

  multi_out <- combine_supp(dataset, suppae)
  expect_equal(multi_out$SUPPVAR1, ae$SUPPVAR1)
  expect_equal(multi_out$SUPPVAR2, ae$SUPPVAR2)
  expect_equal(multi_out$SUPPVAR3, ae$SUPPVAR3)
})

test_that("combine_supp works with different IDVARVAL classes", {
  skip_if_not_installed("pharmaversesdtm")
  expect_equal(
    combine_supp(pharmaversesdtm::ae, pharmaversesdtm::suppae) %>%
      pull(AESEQ),
    pharmaversesdtm::ae %>% pull(AESEQ)
  )
})

test_that("combine_supp works with without QEVAL", {
  skip_if_not_installed("pharmaversesdtm")
  expect_silent(combine_supp(pharmaversesdtm::tr_onco, pharmaversesdtm::supptr_onco))
})

test_that("supp data that does not match the main data will raise a warning but return a dataset", {
  sdtm_suppae_extra <- safetyData::sdtm_suppae
  sdtm_suppae_extra$IDVARVAL[1] <- 99

  expect_warning(
    out <- combine_supp(safetyData::sdtm_ae, sdtm_suppae_extra),
    "Some SUPP records were not merged into the main dataset"
  )
  expect_s3_class(out, "data.frame")
})


test_that("Floating point correction works", {
  fp1 <- 0.1 + 0.1 + 0.1 + 0.1 + 0.1 + 0.1 + 0.1 + 0.1 + 0.1 + 0.1
  sdtm_ae_fp <- safetyData::sdtm_ae %>%
    mutate(AESEQ = case_when(
      AESEQ == 1 ~ fp1,
      TRUE ~ as.double(AESEQ)
    ))
  # correction
  combo_ae <- combine_supp(sdtm_ae_fp, safetyData::sdtm_suppae) %>%
    select(USUBJID, AESEQ, AETRTEM) %>%
    distinct() %>%
    arrange(USUBJID, AESEQ)
  supp_check <- safetyData::sdtm_suppae %>%
    select(USUBJID, AESEQ = IDVARVAL, AETRTEM = QVAL) %>%
    arrange(USUBJID, AESEQ)
  attr(supp_check$AETRTEM, "label") <- "TREATMENT EMERGENT FLAG"
  expect_equal(combo_ae, supp_check)
})

test_that("zero-row supp returns data unchanged with a warning (#45)", {
  expect_warning(
    result <- combine_supp(safetyData::sdtm_ae, safetyData::sdtm_suppae[0, ]),
    regexp = "Zero rows in supp, returning original dataset unchanged"
  )
  expect_equal(result, safetyData::sdtm_ae)
})

test_that("multiple different IDVAR map to the same QNAM works", {
  simple_ae <-
    safetyData::sdtm_ae |>
    filter(USUBJID %in% c("01-701-1015", "01-701-1023"))
  simple_suppae <- safetyData::sdtm_suppae[c(1, 4), ]
  simple_suppae$IDVAR[2] <- "AEDTC"
  simple_suppae$IDVARVAL[2] <- "2012-09-02"
  expect_equal(
    combine_supp(simple_ae, supp = simple_suppae)$AETRTEM,
    structure(c("Y", NA, NA, NA, NA, NA, "Y"), label = "TREATMENT EMERGENT FLAG")
  )

  # Replace the value in error
  simple_suppae <- safetyData::sdtm_suppae[c(1, 4, 7), ]
  simple_suppae$IDVAR[2] <- "AEDTC"
  simple_suppae$IDVARVAL[2] <- "2012-09-02"

  expect_error(
    combine_supp(simple_ae, supp = simple_suppae),
    regexp = "SUPP domain merge failed due to inconsistent key mapping."
  )
})

test_that("label is added in combine_supp() (#71)", {
  simple_ae <-
    safetyData::sdtm_ae |>
    filter(USUBJID %in% c("01-701-1015", "01-701-1023"))
  simple_suppae <- safetyData::sdtm_suppae[c(1, 4), ]
  labelled <- combine_supp(simple_ae, simple_suppae)
  expect_equal(attr(labelled$AETRTEM, "label"), "TREATMENT EMERGENT FLAG")
})

test_that("combine_supp() does not create an IDVARVAL column (#78)", {
  simple_ae <-
    safetyData::sdtm_ae |>
    filter(USUBJID %in% c("01-701-1015", "01-701-1023"))
  simple_suppae <- safetyData::sdtm_suppae[c(1, 4), ]
  noidvarval <- combine_supp(simple_ae, simple_suppae)
  expect_false("IDVARVAL" %in% names(noidvarval))
})

test_that("make_supp_qual handles deprecated dataset_name parameter", {
  load(metacore::metacore_example("pilot_SDTM.rda"))
  ae <- combine_supp(safetyData::sdtm_ae, safetyData::sdtm_suppae)

  # Test that dataset_name is deprecated and shows guidance
  suppressWarnings({
    result <- make_supp_qual(ae, metacore, dataset_name = "AE")
  })

  expect_s3_class(result, "data.frame")
})

test_that("combine_supp handles QNAM not in dataset columns", {
  simple_ae <- safetyData::sdtm_ae[1:5, ]
  simple_suppae <- safetyData::sdtm_suppae[1, ]
  simple_suppae$QNAM <- "NEWCOL" # A new column to add

  # Should successfully add the new column
  result <- combine_supp(simple_ae, simple_suppae)
  expect_true("NEWCOL" %in% names(result))
})

test_that("combine_supp errors when QNAM already exists in dataset", {
  simple_ae <- safetyData::sdtm_ae[1:5, ]
  simple_ae$AETRTEM <- "existing" # Add column that matches QNAM
  simple_suppae <- safetyData::sdtm_suppae[1, ]

  expect_error(
    combine_supp(simple_ae, simple_suppae),
    "Column name conflict detected when combining SUPP data"
  )
})

test_that("combine_supp handles IDVAR not in dataset", {
  simple_ae <- safetyData::sdtm_ae[1:5, ]
  simple_suppae <- safetyData::sdtm_suppae[1, ]
  simple_suppae$IDVAR <- "FAKEIDVAR" # IDVAR that doesn't exist

  expect_warning(
    combine_supp(simple_ae, simple_suppae),
    "The following IDVAR values from the SUPP dataset will not be joined"
  )
})

test_that("combine_supp_by_idvar detects conflicting replacements across IDVARs", {
  simple_ae <- safetyData::sdtm_ae %>%
    filter(USUBJID %in% c("01-701-1015", "01-701-1023")) %>%
    mutate(NEWID = dplyr::row_number())

  # Create supp with same QNAM but different IDVARs that would cause conflicts
  suppae_conflict <- bind_rows(
    data.frame(
      STUDYID = "CDISCPILOT01",
      RDOMAIN = "AE",
      USUBJID = "01-701-1015",
      IDVAR = "AESEQ",
      IDVARVAL = "1",
      QNAM = "TESTVAR",
      QLABEL = "Test Variable",
      QVAL = "ValueA",
      QORIG = "CRF",
      QEVAL = "",
      stringsAsFactors = FALSE
    ),
    data.frame(
      STUDYID = "CDISCPILOT01",
      RDOMAIN = "AE",
      USUBJID = "01-701-1015",
      IDVAR = "NEWID",
      IDVARVAL = "1",
      QNAM = "TESTVAR",
      QLABEL = "Test Variable",
      QVAL = "ValueB", # Different value for same subject/QNAM
      QORIG = "CRF",
      QEVAL = "",
      stringsAsFactors = FALSE
    )
  )

  expect_error(
    combine_supp(simple_ae, suppae_conflict),
    "SUPP domain merge failed due to inconsistent key mapping."
  )
})

test_that("combine_supp: all SUPP rows merge cleanly (#98)", {
  pc <- tibble::tibble(
    STUDYID = "STUDY123",
    DOMAIN = "PC",
    USUBJID = c("01-001", "01-002", "01-003"),
    PCSEQ = c(1, 2, 3),
    PCTESTCD = "CONC",
    PCTEST = "Concentration",
    PCORRES = c("5.1", "7.3", "4.8"),
    PCORRESU = "ng/mL",
    PCSTRESC = c("5.1", "7.3", "4.8"),
    PCSTRESN = c(5.1, 7.3, 4.8),
    PCSTRESU = "ng/mL",
    PCPOS = "PLASMA",
    PCDTC = c("2025-07-15T08:00", "2025-07-15T09:00", "2025-07-15T10:00"),
    VISITNUM = 1,
    VISIT = "Visit 1",
    ARM = "Drug A",
    ACTARM = "Drug A"
  )

  supppc <- tibble::tibble(
    STUDYID  = "STUDY123",
    RDOMAIN  = "PC",
    USUBJID  = c("01-001", "01-002", "01-003"),
    IDVAR    = "PCSEQ",
    IDVARVAL = c(1, 2, 3),
    QNAM     = "PCREASND",
    QLABEL   = "Reason Not Done",
    QVAL     = c("NA", "NA", "NA"),
    QORIG    = "SPONSOR",
    QEVAL    = "INVESTIGATOR"
  )
  expect_no_warning(
    out <- combine_supp(pc, supppc)
  )

  expect_s3_class(out, "data.frame")
  expect_equal(nrow(out), nrow(pc))
  expect_true("PCREASND" %in% names(out))
  expect_equal(
    unname(as.character(out$PCREASND)),
    c("NA", "NA", "NA")
  )
})

test_that("combine_supp: extra SUPP rows that do not match core raise a warning but return a dataset(#98)", {
  pc <- tibble::tibble(
    STUDYID = "STUDY123",
    DOMAIN = "PC",
    USUBJID = c("01-001", "01-002", "01-003"),
    PCSEQ = c(1, 2, 3),
    PCTESTCD = "CONC",
    PCTEST = "Concentration",
    PCORRES = c("5.1", "7.3", "4.8"),
    PCORRESU = "ng/mL",
    PCSTRESC = c("5.1", "7.3", "4.8"),
    PCSTRESN = c(5.1, 7.3, 4.8),
    PCSTRESU = "ng/mL",
    PCPOS = "PLASMA",
    PCDTC = c("2025-07-15T08:00", "2025-07-15T09:00", "2025-07-15T10:00"),
    VISITNUM = 1,
    VISIT = "Visit 1",
    ARM = "Drug A",
    ACTARM = "Drug A"
  )

  supppc <- tibble::tibble(
    STUDYID  = "STUDY123",
    RDOMAIN  = "PC",
    USUBJID  = c("01-001", "01-002", "01-003"),
    IDVAR    = "PCSEQ",
    IDVARVAL = c(1, 2, 3),
    QNAM     = "PCREASND",
    QLABEL   = "Reason Not Done",
    QVAL     = c("NA", "NA", "NA"),
    QORIG    = "SPONSOR",
    QEVAL    = "INVESTIGATOR"
  )
  supppc_extra <- dplyr::bind_rows(
    supppc,
    dplyr::mutate(supppc[3, ], IDVARVAL = 99),
    dplyr::mutate(supppc[3, ], IDVARVAL = 101)
  )

  expect_warning(
    out <- combine_supp(pc, supppc_extra),
    "Some SUPP records were not merged into the main dataset"
  )
  expect_s3_class(out, "data.frame")
  expect_equal(nrow(out), nrow(pc))
  expect_true("PCREASND" %in% names(out))
  expect_equal(
    unname(as.character(out$PCREASND)),
    c("NA", "NA", "NA")
  )
  expect_false(any(out$PCSEQ %in% c(99, 101)))
})

test_that("build_qnam verbose parameter", {
  # Create simple test data with a column that will be used as QNAM
  ae <- safetyData::sdtm_ae %>%
    head(10) %>%
    mutate(TESTVAR = c("", "", "Y", "Y", "N", "", "Y", "N", "Y", "")) # Some empty strings

  # Test verbose = "message" (default) - should show message about empty QVAL
  expect_message(
    build_qnam(
      dataset = ae,
      qnam = "TESTVAR",
      qlabel = "Test Variable",
      idvar = "AESEQ",
      qeval = "INVESTIGATOR",
      qorig = "CRF",
      verbose = "message"
    ),
    "Empty QVAL rows removed for QNAM = TESTVAR"
  )

  # Test verbose = "warn" - suppress messages
  expect_silent(
    result_warn <- build_qnam(
      dataset = ae,
      qnam = "TESTVAR",
      qlabel = "Test Variable",
      idvar = "AESEQ",
      qeval = "INVESTIGATOR",
      qorig = "CRF",
      verbose = "warn"
    )
  )

  # Test verbose = "silent" - suppress all output
  expect_silent(
    result_silent <- build_qnam(
      dataset = ae,
      qnam = "TESTVAR",
      qlabel = "Test Variable",
      idvar = "AESEQ",
      qeval = "INVESTIGATOR",
      qorig = "CRF",
      verbose = "silent"
    )
  )

  # Verify all verbose levels return same result
  result_message <- suppressMessages(
    build_qnam(
      dataset = ae,
      qnam = "TESTVAR",
      qlabel = "Test Variable",
      idvar = "AESEQ",
      qeval = "INVESTIGATOR",
      qorig = "CRF",
      verbose = "message"
    )
  )

  expect_equal(result_message, result_warn)
  expect_equal(result_message, result_silent)

  # Verify empty strings were actually removed
  expect_false("" %in% result_message$QVAL)

  # Test invalid verbose value
  expect_error(
    build_qnam(
      dataset = ae,
      qnam = "TESTVAR",
      qlabel = "Test Variable",
      idvar = "AESEQ",
      qeval = "INVESTIGATOR",
      qorig = "CRF",
      verbose = "invalid"
    ),
    "should be one of: message, warn, silent"
  )
})

test_that("combine_supp throws clean errors and warnings", {
  ae <- data.frame(
    STUDYID = c("ABC123", "ABC123", "ABC123", "ABC123", "ABC123"),
    DOMAIN = c("AE", "AE", "AE", "AE", "AE"),
    USUBJID = c("ABC123-001", "ABC123-001", "ABC123-002", "ABC123-002", "ABC123-003"),
    AESEQ = c(1, 2, 1, 2, 1),
    AESPID = c("AE01", "AE02", "AE01", "AE02", "AE01"),
    AEGRPID = c(NA, "GRP-A", NA, "GRP-B", "GRP-A"),
    AETERM = c("Headache", "Nausea", "Dizziness", "Vomiting", "Rash")
  )

  suppae <- data.frame(
    STUDYID = c("ABC123", "ABC123", "ABC123", "ABC123", "ABC123", "ABC123"),
    RDOMAIN = c("AE", "AE", "AE", "AE", "AE", "AE"),
    USUBJID = c("ABC123-001", "ABC123-001", "ABC123-001", "ABC123-002", "ABC123-002", "ABC123-003"),
    IDVAR = c("AESEQ", "AESEQ", "AESPID", "AESPID", "AEGRPID", "AEGRPID"),
    IDVARVAL = c("1", "2", "AE01", "AE02", "GRP-B", "GRP-A"),
    QNAM = c("AELAT", "AETOXGR", "AEREL", "AEPATT", "AESEV2", "AESOC2"),
    QLABEL = c("Laterality", "Toxicity Grade", "Relationship", "Pattern", "Severity (Alt)", "SOC (Alt)"),
    QVAL = c("LEFT", "2", "RELATED", "INTERMITTENT", "MODERATE", "SKIN DISORDERS"),
    QORIG = c("CRF", "CRF", "CRF", "CRF", "DERIVED", "DERIVED"),
    QEVAL = c(NA, NA, "INVESTIGATOR", NA, NA, NA)
  )

  # Missing variables in supp, does not comply with CDISC structure
  expect_error(
    combine_supp(
      dataset = ae,
      supp = suppae |> select(-RDOMAIN, -IDVAR, -QNAM)
    ),
    "Supplemental Qualifier dataset does not comply with CDISC SDTM structure."
  )

  # Additional variables in supp, does not comply with CDISC structure
  expect_error(
    combine_supp(
      dataset = ae,
      supp = suppae |> mutate(EXT = NA_character_)
    ),
    "Supplemental Qualifier dataset does not comply with CDISC SDTM structure."
  )

  # IDVAR values in the supp are not in the main dataset and will not be joined
  ae_01 <- ae[, !names(ae) %in% unique(suppae$IDVAR)]
  expect_warning(
    combine_supp(
      dataset = ae_01,
      supp = suppae
    ),
    "The following IDVAR values from the SUPP dataset will not be joined"
  )

  # Core SDTM variables are missing from the main dataset
  ae_02 <- ae[, !names(ae) %in% c("STUDYID", "DOMAIN", "USUBJID")]
  expect_error(
    combine_supp(
      dataset = ae_02,
      supp = suppae
    ),
    "Core SDTM variables are missing from the dataset"
  )
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build.R
\name{add_variables}
\alias{add_variables}
\title{Add Missing Variables}
\usage{
add_variables(dataset, metacore, dataset_name = deprecated())
}
\arguments{
\item{dataset}{Dataset to add columns to. If all variables are present no
columns will be added.}

\item{metacore}{metacore object that only contains the specifications for the
dataset of interest.}

\item{dataset_name}{Optional string to specify the dataset. This is only
needed if the metacore object provided hasn't already been subsetted.\cr
Note: Deprecated in version 0.2.0. The \code{dataset_name} argument will be removed
in a future release. Please use \code{metacore::select_dataset} to subset the
\code{metacore} object to obtain metadata for a single dataset.}
}
\value{
The given dataset with any additional columns added
}
\description{
This function adds in missing columns according to the type set in the
metacore object. All values in the new columns will be missing, but typed
correctly. If unable to recognize the type in the metacore object will return
a logical type.
}
\examples{
library(metacore)
library(haven)
library(dplyr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
data <- read_xpt(metatools_example("adsl.xpt")) \%>\%
  select(-TRTSDT, -TRT01P, -TRT01PN)
add_variables(data, spec)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build.R
\name{drop_unspec_vars}
\alias{drop_unspec_vars}
\title{Drop Unspecified Variables}
\usage{
drop_unspec_vars(
  dataset,
  metacore,
  dataset_name = deprecated(),
  verbose = c("message", "warn", "silent")
)
}
\arguments{
\item{dataset}{Dataset to change}

\item{metacore}{metacore object that only contains the specifications for the
dataset of interest.}

\item{dataset_name}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Optional string to specify
the dataset. This is only needed if the metacore object provided hasn't already
been subsetted.\cr
Note: Deprecated in version 0.2.0. The \code{dataset_name} argument will be removed
in a future release. Please use \code{metacore::select_dataset} to subset the
\code{metacore} object to obtain metadata for a single dataset.}

\item{verbose}{Character string controlling message verbosity. One of:
\describe{
\item{\code{"message"}}{Show both warnings and messages (default)}
\item{\code{"warn"}}{Show warnings but suppress messages}
\item{\code{"silent"}}{Suppress all warnings and messages}
}}
}
\value{
Dataset with only specified columns
}
\description{
This function drops all unspecified variables.
}
\examples{
library(metacore)
library(haven)
library(dplyr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
data <- read_xpt(metatools_example("adsl.xpt")) \%>\%
  select(USUBJID, SITEID) \%>\%
  mutate(foo = "Hello")
drop_unspec_vars(data, spec)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_ct_data}
\alias{check_ct_data}
\title{Check Control Terminology for a Dataset}
\usage{
check_ct_data(
  data,
  metacore,
  na_acceptable = NULL,
  omit_vars = NULL,
  verbose = "message"
)
}
\arguments{
\item{data}{Dataset to check}

\item{metacore}{metacore object that contains the specifications for the
dataset of interest. If any variable has different codelists for different
datasets the metacore object will need to be subsetted using
\code{select_dataset} from the metacore package.}

\item{na_acceptable}{\code{logical} value or \code{character} vector, set to \code{NULL} by default.
\code{NULL} sets the acceptability of missing values based on if the core for
the variable is "Required" in the \code{metacore} object. If set to \code{TRUE} then will
pass check if values are in the control terminology or are missing. If set
to \code{FALSE} then NA will not be acceptable. If set to a \code{character} vector then
only the specified variables may contain NA values.}

\item{omit_vars}{\code{character} vector indicating which variables should be skipped
when doing the controlled terminology checks. Internally, \code{omit_vars} is
evaluated before \code{na_acceptable}.}

\item{verbose}{\code{character} string controlling the verbosity of the output.
Possible values are \code{"message"} (for general information and success messages)
and \code{"warn"} (for warnings). Partial matching is allowed.
\strong{Important}: \code{"silent"} is explicitly \strong{not} a valid option for \code{verbose} in
this function. The primary purpose of \code{check_ct_data} is to identify and
warn the user about non-compliant or problematic control terminology. Allowing
the suppression of these warnings would bypass the function's intent and could
lead to unnoticed data quality issues. If \code{verbose = "silent"} is provided, it will
be coerced to \code{"message"} with a warning.}
}
\value{
Given data if all columns pass. It will issue a warning otherwise.
}
\description{
This function checks that all columns in the dataset only contains the
control terminology as defined by the metacore specification.
}
\examples{
library(haven)
library(metacore)
library(magrittr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL", quiet = TRUE)
data <- read_xpt(metatools_example("adsl.xpt"))

check_ct_data(data, spec, omit_vars = c("AGEGR2", "AGEGR2N"))
\dontrun{
# These examples produce errors:
check_ct_data(data, spec, na_acceptable = FALSE)
check_ct_data(data, spec, na_acceptable = FALSE, omit_vars = "DISCONFL")
check_ct_data(data, spec, na_acceptable = c("DSRAEFL", "DCSREAS"), omit_vars = "DISCONFL")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build.R
\name{build_from_derived}
\alias{build_from_derived}
\title{Build a dataset from derived}
\usage{
build_from_derived(
  metacore,
  ds_list,
  dataset_name = deprecated(),
  predecessor_only = TRUE,
  keep = FALSE,
  verbose = c("message", "warn", "silent")
)
}
\arguments{
\item{metacore}{metacore object that contains the specifications for the
dataset of interest.}

\item{ds_list}{Named list of datasets that are needed to build the from. If
the list is unnamed,then it will use the names of the objects.}

\item{dataset_name}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Optional string to
specify the dataset that is being built. This is only needed if the metacore
object provided hasn't already been subsetted.\cr
Note: Deprecated in version 0.2.0. The \code{dataset_name} argument will be removed
in a future release. Please use \code{metacore::select_dataset} to subset the
\code{metacore} object to obtain metadata for a single dataset.}

\item{predecessor_only}{By default \code{TRUE}, so only variables with the origin
of 'Predecessor' will be used. If \code{FALSE} any derivation matching the
dataset.variable will be used.}

\item{keep}{String to determine which columns from the original datasets
should be kept
\itemize{
\item "FALSE" (default):  only columns that are also present in the ADaM
specification are kept in the output.
\item "ALL":              all original columns are carried through to the
ADaM, including those that have been renamed.
e.g. if DM.ARM is a predecessor to DM.TRT01P,
both ARM and TRT01P will be present as columns
in the ADaM output.
\item "PREREQUISITE":     columns are retained if they are required for future
derivations in the specification. Additional
prerequisite columns are identified as columns
that appear in the 'derivation' column of the
metacore object in the format "DATASET.VARIABLE",
but not as direct predecessors. Predecessors are
defined as columns where the derivation is a 1:1
copy of a column in a source dataset.
e.g. derivation = "VS.VSTESTCD" is a predecessor,
while derivation = "Value of VS.VSSTRESN where
VS.VSTESTCD == 'Heart Rate'" contains both
VS.VSTESTCD and VS.VSSTRESN as prerequisites, and
these columns will be kept through to the ADaM.
}}

\item{verbose}{Character string controlling message verbosity. One of:
\describe{
\item{\code{"message"}}{Show both warnings and messages (default)}
\item{\code{"warn"}}{Show warnings but suppress messages}
\item{\code{"silent"}}{Suppress all warnings and messages}
}}
}
\value{
dataset
}
\description{
This function builds a dataset out of the columns that just need to be pulled
through. So any variable that has a derivation in the format of
'dataset.variable' will be pulled through to create the new dataset. When
there are multiple datasets present, they will be joined by the shared
\code{key_seq} variables. These columns are often called 'Predecessors' in ADaM,
but this is not universal so that is optional to specify.
}
\examples{
library(metacore)
library(haven)
library(magrittr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
ds_list <- list(DM = read_xpt(metatools_example("dm.xpt")))
build_from_derived(spec, ds_list, predecessor_only = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{remove_labels}
\alias{remove_labels}
\title{Remove labels to multiple variables on a data frame}
\usage{
remove_labels(data)
}
\arguments{
\item{data}{A data.frame or tibble}
}
\value{
data with variable labels applied
}
\description{
This function allows a user to removes all labels to a dataframe at once.
}
\examples{
library(haven)
data <- read_xpt(metatools_example("adsl.xpt"))
remove_labels(data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{set_variable_labels}
\alias{set_variable_labels}
\title{Apply labels to a data frame using a metacore object}
\usage{
set_variable_labels(
  data,
  metacore,
  dataset_name = deprecated(),
  verbose = c("message", "warn", "silent")
)
}
\arguments{
\item{data}{A dataframe or tibble upon which labels will be applied}

\item{metacore}{metacore object that contains the specifications for the
dataset of interest.}

\item{dataset_name}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Optional string to
specify the dataset that is being built. This is only needed if the metacore
object provided hasn't already been subsetted.\cr
Note: Deprecated in version 0.2.0. The \code{dataset_name} argument will be removed
in a future release. Please use \code{metacore::select_dataset} to subset the
\code{metacore} object to obtain metadata for a single dataset.}

\item{verbose}{Character string controlling message verbosity. One of:
\describe{
\item{\code{"message"}}{Show both warnings and messages (default)}
\item{\code{"warn"}}{Show warnings but suppress messages}
\item{\code{"silent"}}{Suppress all warnings and messages}
}}
}
\value{
Dataframe with labels applied
}
\description{
This function leverages metadata available in a metacore object to apply
labels to a data frame.
}
\examples{

mc <- metacore::spec_to_metacore(
  metacore::metacore_example("p21_mock.xlsx"),
  quiet = TRUE
)
dm <- haven::read_xpt(metatools_example("dm.xpt"))
set_variable_labels(dm, mc, dataset_name = "DM")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/codelists.R
\name{create_cat_var}
\alias{create_cat_var}
\title{Create Categorical Variable from Codelist}
\usage{
create_cat_var(
  data,
  metacore,
  ref_var,
  grp_var,
  num_grp_var = NULL,
  create_from_decode = FALSE,
  strict = TRUE
)
}
\arguments{
\item{data}{Dataset with reference variable in it}

\item{metacore}{A metacore object to get the codelist from. If the
variable has different codelists for different datasets the metacore object
will need to be subsetted using \code{select_dataset} from the metacore package.}

\item{ref_var}{Name of variable to be used as the reference i.e AGE when
creating AGEGR1}

\item{grp_var}{Name of the new grouped variable}

\item{num_grp_var}{Name of the new numeric decode for the grouped variable.
This is optional if no value given no variable will be created}

\item{create_from_decode}{Sets the \code{decode} column of the codelist as the column
from which the variable will be created. By default the column is \code{code}.}

\item{strict}{A logical value indicating whether to perform strict checking
against the codelist. If \code{TRUE} will issue a warning if values in the \code{ref_var}
column do not fit into the group definitions for the codelist in \code{grp_var}.
If \code{FALSE} no warning is issued and values not defined by the codelist will
likely result in \code{NA} results.}
}
\value{
dataset with new column added
}
\description{
Using the grouping from either the \code{decode_var} or \code{code_var} and a reference
variable (\code{ref_var}) it will create a categorical variable and the numeric
version of that categorical variable.
}
\examples{
library(metacore)
library(haven)
library(dplyr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
dm <- read_xpt(metatools_example("dm.xpt")) \%>\%
  select(USUBJID, AGE)
# Grouping Column Only
create_cat_var(dm, spec, AGE, AGEGR1)
# Grouping Column and Numeric Decode
create_cat_var(dm, spec, AGE, AGEGR1, AGEGR1N)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_ct_col}
\alias{check_ct_col}
\title{Check Control Terminology for a Single Column}
\usage{
check_ct_col(
  data,
  metacore,
  var,
  na_acceptable = NULL,
  verbose = "message",
  .internal = FALSE
)
}
\arguments{
\item{data}{Data to check}

\item{metacore}{A metacore object to get the codelist from. If the variable
has different codelists for different datasets the metacore object will
need to be subsetted using \code{select_dataset} from the metacore package.}

\item{var}{Name of variable to check}

\item{na_acceptable}{Logical value, set to \code{NULL} by default, so the
acceptability of missing values is based on if the core for the variable is
"Required" in the \code{metacore} object. If set to \code{TRUE} then will pass check
if values are in the control terminology or are missing. If set to
\code{FALSE}then NA will not be acceptable.}

\item{verbose}{\code{character} string controlling the verbosity of the output.
Possible values are \code{"message"} (for general information and success messages)
and \code{"warn"} (for warnings). Partial matching is allowed.
\strong{Important}: \code{"silent"} is explicitly \strong{not} a valid option for \code{verbose} in
this function. The primary purpose of \code{check_ct_data} is to identify and
warn the user about non-compliant or problematic control terminology. Allowing
the suppression of these warnings would bypass the function's intent and could
lead to unnoticed data quality issues. If \code{verbose = "silent"} is provided, it will
be coerced to \code{"message"} with a warning.}

\item{.internal}{Logical value indicating whether the function is being
called internally by another package function. If \code{TRUE}, the function
suppresses user-facing messages and instead returns a logical indicator
of whether any controlled terminology violations were detected. This
argument is intended for internal use only and should not be set by
end users.}
}
\value{
Given data if column only contains control terms. If not, will error
given the values which should not be in the column
}
\description{
This function checks the column in the dataset only contains the control
terminology as defined by the metacore specification
}
\examples{
library(metacore)
library(haven)
library(magrittr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
data <- read_xpt(metatools_example("adsl.xpt"))
check_ct_col(data, spec, TRT01PN)
check_ct_col(data, spec, "TRT01PN")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supp.R
\name{combine_supp}
\alias{combine_supp}
\title{Combine the Domain and Supplemental Qualifier}
\usage{
combine_supp(dataset, supp)
}
\arguments{
\item{dataset}{Domain dataset}

\item{supp}{Supplemental Qualifier dataset}
}
\value{
a dataset with the supp variables added to it
}
\description{
Combine the Domain and Supplemental Qualifier
}
\examples{
library(safetyData)
library(tibble)
combine_supp(sdtm_ae, sdtm_suppae) \%>\% as_tibble()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{check_variables}
\alias{check_variables}
\title{Check Variable Names}
\usage{
check_variables(data, metacore, dataset_name = deprecated(), strict = FALSE)
}
\arguments{
\item{data}{Dataset to check}

\item{metacore}{metacore object that only contains the specifications for the
dataset of interest.}

\item{dataset_name}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Optional string to
specify the dataset. This is only needed if the metacore object provided
hasn't already been subsetted.\cr
Note: Deprecated in version 0.2.0. The \code{dataset_name} argument will be removed
in a future release. Please use \code{metacore::select_dataset} to subset the
\code{metacore} object to obtain metadata for a single dataset.}

\item{strict}{A logical value indicating whether to perform strict
validation on the input dataset. If \code{TRUE} (default), errors will be raised
if validation fails. If \code{FALSE}, warnings will be issued instead, allowing
the function execution to continue event with invalid data.}
}
\value{
message if the dataset matches the specification and the dataset, and error otherwise
}
\description{
This function checks the variables in the dataset against the variables
defined in the metacore specifications. If everything matches the function
will print a message stating everything is as expected. If there
are additional or missing variables an error will explain the discrepancies
}
\examples{
library(haven)
library(metacore)
library(magrittr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
data <- read_xpt(metatools_example("adsl.xpt"))
check_variables(data, spec)
data["DUMMY_COL"] <- NA
check_variables(data, spec, strict = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checks.R
\name{get_bad_ct}
\alias{get_bad_ct}
\title{Gets vector of control terminology which should be there}
\usage{
get_bad_ct(data, metacore, var, na_acceptable = NULL)
}
\arguments{
\item{data}{Data to check}

\item{metacore}{A metacore object to get the codelist from. If the variable
has different codelists for different datasets the metacore object will
need to be subsetted using \code{select_dataset} from the metacore package.}

\item{var}{Name of variable to check}

\item{na_acceptable}{Logical value, set to \code{NULL} by default, so the
acceptability of missing values is based on if the core for the variable is
"Required" in the \code{metacore} object. If set to \code{TRUE} then will pass check
if values are in the control terminology or are missing. If set to
\code{FALSE} then NA will not be acceptable.}
}
\value{
vector
}
\description{
This function checks the column in the dataset only contains the control
terminology as defined by the metacore specification. It will return all
values not found in the control terminology
}
\examples{
library(haven)
library(metacore)
library(magrittr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
data <- read_xpt(metatools_example("adsl.xpt"))
get_bad_ct(data, spec, "DCSREAS")
get_bad_ct(data, spec, "DCSREAS", na_acceptable = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sort.R
\name{sort_by_key}
\alias{sort_by_key}
\title{Sort Rows by Key Sequence}
\usage{
sort_by_key(data, metacore, dataset_name = deprecated())
}
\arguments{
\item{data}{Dataset to sort}

\item{metacore}{metacore object that contains the specifications for the
dataset of interest.}

\item{dataset_name}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Optional string to
specify the dataset that is being built. This is only needed if the metacore
object provided hasn't already been subsetted.\cr
Note: Deprecated in version 0.2.0. The \code{dataset_name} argument will be removed
in a future release. Please use \code{metacore::select_dataset} to subset the
\code{metacore} object to obtain metadata for a single dataset.}
}
\value{
dataset with ordered columns
}
\description{
This function sorts the dataset according to the key sequence found in the
metacore object.
}
\examples{
library(metacore)
library(haven)
library(magrittr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
data <- read_xpt(metatools_example("adsl.xpt"))
sort_by_key(data, spec)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{add_labels}
\alias{add_labels}
\title{Apply labels to multiple variables on a data frame}
\usage{
add_labels(data, ...)
}
\arguments{
\item{data}{A data.frame or tibble}

\item{...}{Named parameters in the form of variable = 'label'}
}
\value{
data with variable labels applied
}
\description{
This function allows a user to apply several labels to a dataframe at once.
}
\examples{
add_labels(
  mtcars,
  mpg = "Miles Per Gallon",
  cyl = "Cylinders"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supp.R
\name{build_qnam}
\alias{build_qnam}
\title{Build the observations for a single QNAM}
\usage{
build_qnam(
  dataset,
  qnam,
  qlabel,
  idvar,
  qeval,
  qorig,
  verbose = c("message", "warn", "silent")
)
}
\arguments{
\item{dataset}{Input dataset}

\item{qnam}{QNAM value}

\item{qlabel}{QLABEL value}

\item{idvar}{IDVAR variable name (provided as a string)}

\item{qeval}{QEVAL value to be populated for this QNAM}

\item{qorig}{QORIG value to be populated for this QNAM}

\item{verbose}{Character string controlling message verbosity. One of:
\describe{
\item{\code{"message"}}{Show both warnings and messages (default)}
\item{\code{"warn"}}{Show warnings but suppress messages}
\item{\code{"silent"}}{Suppress all warnings and messages}
}}
}
\value{
Observations structured in SUPP format
}
\description{
Build the observations for a single QNAM
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/supp.R
\name{make_supp_qual}
\alias{make_supp_qual}
\title{Make Supplemental Qualifier}
\usage{
make_supp_qual(dataset, metacore, dataset_name = deprecated())
}
\arguments{
\item{dataset}{dataset the supp will be pulled from}

\item{metacore}{A subsetted metacore object to get the supp information from.
If not already subsetted then a \code{dataset_name} will need to be provided}

\item{dataset_name}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Optional string to
specify the dataset that is being built. This is only needed if the metacore
object provided hasn't already been subsetted.\cr
Note: Deprecated in version 0.2.0. The \code{dataset_name} argument will be removed
in a future release. Please use \code{metacore::select_dataset} to subset the
\code{metacore} object to obtain metadata for a single dataset.}
}
\value{
a CDISC formatted SUPP dataset
}
\description{
Make Supplemental Qualifier
}
\examples{

library(metacore)
library(safetyData)
library(tibble)
load(metacore_example("pilot_SDTM.rda"))
spec <- metacore \%>\% select_dataset("AE")
ae <- combine_supp(sdtm_ae, sdtm_suppae)
make_supp_qual(ae, spec) \%>\% as_tibble()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metatools-package.R
\docType{package}
\name{metatools-package}
\alias{metatools}
\alias{metatools-package}
\title{metatools: Enable the Use of 'metacore' to Help Create and Check Dataset}
\description{
Uses the metadata information stored in 'metacore' objects to check and build metadata associated columns.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/pharmaverse/metatools}
  \item \url{https://pharmaverse.github.io/metatools/}
  \item Report bugs at \url{https://github.com/pharmaverse/metatools/issues}
}

}
\author{
\strong{Maintainer}: Liam Hobby \email{liam.f.hobby@gsk.com}

Authors:
\itemize{
  \item Christina Fillmore \email{christina.e.fillmore@gsk.com} (\href{https://orcid.org/0000-0003-0595-2302}{ORCID})
  \item Bill Denney
  \item Mike Stackhouse \email{mike.stackhouse@atorusresearch.com} (\href{https://orcid.org/0000-0001-6030-723X}{ORCID})
  \item Jana Stoilova \email{jana.stoilova@roche.com}
  \item Tamara Senior \email{tamara.senior@roche.com}
}

Other contributors:
\itemize{
  \item GlaxoSmithKline LLC [copyright holder, funder]
  \item F. Hoffmann-La Roche AG [copyright holder, funder]
  \item Atorus Research LLC [copyright holder, funder]
}

}
\concept{internal}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/codelists.R
\name{create_var_from_codelist}
\alias{create_var_from_codelist}
\title{Create Variable from Codelist}
\usage{
create_var_from_codelist(
  data,
  metacore,
  input_var,
  out_var,
  codelist = NULL,
  decode_to_code = TRUE,
  strict = TRUE
)
}
\arguments{
\item{data}{Dataset that contains the input variable}

\item{metacore}{A metacore object to get the codelist from. This should be a
subsetted metacore object (of subclass \code{DatasetMeta}) created using
\code{metacore::select_dataset}.}

\item{input_var}{Name of the variable that will be translated for the new
column}

\item{out_var}{Name of the output variable. Note: Unless a codelist is provided
the grouping will always be from the code of the codelist associates with
\code{out_var}.}

\item{codelist}{Optional argument to supply a codelist. Must be a data.frame
with \code{code} and \code{decode} columns such as those created by the function
\code{metacore::get_control_term}. If no codelist is provided the codelist
associated with the column supplied to \code{out_var} will be used. By default
\code{codelist} is \code{NULL}.}

\item{decode_to_code}{Direction of the translation. Default value is \code{TRUE},
i.e., assumes the  \code{input_var} is the decode column of the codelist.
Set to \code{FALSE} if the \code{input_var} is the code column of the codelist.}

\item{strict}{A logical value indicating whether to perform strict checking
against the codelist. If \code{TRUE} will issue a warning if values in the \code{input_var}
column are not present in the codelist. If \code{FALSE} no warning is issued and
values not present in the codelist will likely result in \code{NA} results.}
}
\value{
Dataset with a new column added
}
\description{
This functions uses code/decode pairs from a metacore object to create new
variables in the data
}
\examples{
library(metacore)
library(tibble)
data <- tribble(
  ~USUBJID, ~VAR1, ~VAR2,
  1, "M", "Male",
  2, "F", "Female",
  3, "F", "Female",
  4, "U", "Unknown",
  5, "M", "Male",
)
spec <- spec_to_metacore(metacore_example("p21_mock.xlsx"), quiet = TRUE)
dm_spec <- select_dataset(spec, "DM", quiet = TRUE)
create_var_from_codelist(data, dm_spec, VAR2, SEX)
create_var_from_codelist(data, dm_spec, "VAR2", "SEX")
create_var_from_codelist(data, dm_spec, VAR1, SEX, decode_to_code = FALSE)

# Example providing a custom codelist
# This example also reverses the direction of translation
load(metacore_example("pilot_ADaM.rda"))
adlb_spec <- select_dataset(metacore, "ADLBC", quiet = TRUE)
adlb <- tibble(PARAMCD = c("ALB", "ALP", "ALT", "AST", "BILI", "BUN"))
create_var_from_codelist(
  adlb,
  adlb_spec,
  PARAMCD,
  PARAM,
  codelist = get_control_term(adlb_spec, PARAMCD),
  decode_to_code = FALSE,
  strict = FALSE
)

\dontrun{
# Example expecting warning where `strict` == `TRUE`
adlb <- tibble(PARAMCD = c("ALB", "ALP", "ALT", "AST", "BILI", "BUN", "DUMMY1", "DUMMY2"))
create_var_from_codelist(
  adlb,
  adlb_spec,
  PARAMCD,
  PARAM,
  codelist = get_control_term(adlb_spec, PARAMCD),
  decode_to_code = FALSE,
  strict = TRUE
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/codelists.R
\name{create_subgrps}
\alias{create_subgrps}
\title{Create Subgroups}
\usage{
create_subgrps(ref_vec, grp_defs, grp_labs = NULL)
}
\arguments{
\item{ref_vec}{Vector of numeric values}

\item{grp_defs}{Vector of strings with groupings defined. Format must be
either: <00, >=00, 00-00, or  00-<00}

\item{grp_labs}{Vector of strings with labels defined. The labels correspond
to the associated \code{grp_defs}. i.e., "12-17" may translate to "12-17 years".
If no \code{grp_labs} specified then \code{grp_defs} will be used.}
}
\value{
Character vector of the values in the subgroups
}
\description{
Create Subgroups
}
\examples{
create_subgrps(c(1:10), c("<2", "2-5", ">5"))
create_subgrps(c(1:10), c("<=2", ">2-5", ">5"))
create_subgrps(c(1:10), c("<2", "2-<5", ">=5"))
create_subgrps(c(1:10), c("<2", "2-<5", ">=5"), c("<2 years", "2-5 years", ">=5 years"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/codelists.R
\name{convert_var_to_fct}
\alias{convert_var_to_fct}
\title{Convert Variable to Factor with Levels Set by Control Terms}
\usage{
convert_var_to_fct(data, metacore, var)
}
\arguments{
\item{data}{A dataset containing the variable to be modified}

\item{metacore}{A metacore object to get the codelist from. If the
variable has different codelists for different datasets the metacore object
will need to be subsetted using \code{select_dataset} from the metacore package}

\item{var}{Name of variable to change}
}
\value{
Dataset with variable changed to a factor
}
\description{
This functions takes a dataset, a metacore object and a variable name. Then
looks at the metacore object for the control terms for the given variable and
uses that to convert the variable to a factor with those levels. If the
control terminology is a code list, the code column will be used. The
function fails if the control terminology is an external library
}
\examples{
library(metacore)
library(haven)
library(dplyr)
load(metacore_example("pilot_ADaM.rda"))
spec <- metacore \%>\% select_dataset("ADSL")
dm <- read_xpt(metatools_example("dm.xpt")) \%>\%
  select(USUBJID, SEX, ARM)
# Variable with codelist control terms
convert_var_to_fct(dm, spec, SEX)
# Variable with permitted value control terms
convert_var_to_fct(dm, spec, ARM)
}
